import numpy as np
from numpy.linalg import inv, det
from scipy.stats import norm
from scipy.linalg import sqrtm
from sklearn.cluster import KMeans
from scipy.optimize import minimize
import time
from fmvmm.mixtures._base import BaseMixture
from fmvmm.utils.utils_mixture import (mixture_clusters)
from fmvmm.mixsmsn.information_matrix_smsn import info_matrix, standard_errors_from_info_matrix

def dmvnorm(X, mu, Sigma):
    """
    Multivariate normal density for each row of X.
    X : (n, p)
    mu : (p,) array
    Sigma : (p, p) covariance matrix
    Returns: (n,) array of PDF values
    """
    X = np.asarray(X)
    mu = np.asarray(mu)
    n, p = X.shape

    # Ensure invertibility, add small regularization if needed
    det_Sigma = np.linalg.det(Sigma)
    if det_Sigma < 1e-300:
        # fallback or regularize
        Sigma = Sigma + 1e-6*np.eye(p)
        det_Sigma = np.linalg.det(Sigma)

    inv_Sigma = np.linalg.inv(Sigma)

    diff = X - mu
    quad_form = np.sum(diff @ inv_Sigma * diff, axis=1)  # (n,)

    norm_const = (2.0*np.pi)**(p/2.0) * np.sqrt(det_Sigma)
    pdf_vals = np.exp(-0.5 * quad_form) / norm_const
    return pdf_vals

def dmvSNC(X, mu, Sigma, lam, nu):
    """
    Multivariate Skew Contaminated Normal density for each row in X.
    
    X: (n, p)
    mu: (p,)
    Sigma: (p, p)
    lam: (p,)  -- skewness vector
    nu: (2,)   -- [nu1, nu2], each in (0, 1) typically

    Returns: (n,) array of PDF values
    """
    X = np.asarray(X)
    mu = np.asarray(mu)
    lam = np.asarray(lam)
    n, p = X.shape
    nu1, nu2 = nu[0], nu[1]

    # Precompute Sigma^{-1/2}:
    # Protect against numerical issues if Sigma is near-singular
    det_Sigma = np.linalg.det(Sigma)
    if det_Sigma < 1e-300:
        Sigma = Sigma + 1e-6*np.eye(p)
    sqrt_Sigma = sqrtm(Sigma)
    inv_sqrt_Sigma = np.linalg.inv(sqrt_Sigma)

    # factor = lam^T * inv_sqrt_Sigma  => shape (1, p)
    # We'll apply to (X - mu) row-wise:
    factor = lam @ inv_sqrt_Sigma  # shape (p,) dot (p,p) => shape (p,)
    
    # A1[i] = sqrt(nu2) * factor . (X[i]-mu)
    # A2[i] = factor . (X[i]-mu)
    diff = (X - mu)
    A2 = np.einsum('j,ij->i', factor, diff)     # shape (n,)
    A1 = np.sqrt(nu2)*A2                        # shape (n,)

    # Evaluate the two normal densities:
    # 1) dmvnorm with Sigma/nu2
    dmvnorm_part1 = dmvnorm(X, mu, Sigma/nu2)
    # 2) dmvnorm with Sigma
    dmvnorm_part2 = dmvnorm(X, mu, Sigma)

    # Combine with cdf:
    term1 = nu1 * dmvnorm_part1 * norm.cdf(A1)
    term2 = (1.0 - nu1) * dmvnorm_part2 * norm.cdf(A2)

    dens = 2.0*(term1 + term2)
    return dens

def d_mixedmvSNC(X, pi_list, mu_list, Sigma_list, lam_list, nu):
    """
    Mixture of g Skew-CN components, all sharing the same nu=[nu1, nu2].

    pi_list: (g,) mixing proportions
    mu_list, Sigma_list, lam_list: each is a list of length g
    nu: (2,) array
    Returns: (n,) vector of mixture densities
    """
    g = len(pi_list)
    total = np.zeros(X.shape[0])
    for j in range(g):
        val_j = dmvSNC(X, mu_list[j], Sigma_list[j], lam_list[j], nu)
        total += pi_list[j]*val_j
    return total

def k_means_init_skewcn(data, k):
    """
    Perform K-means initialization for a Skew Contaminated Normal mixture.
    - data : (n, p) array
    - k    : number of clusters

    Returns
    -------
    pi_init  : 1D array of length k
    alpha_init : list of length k, where
                 alpha_init[j] = (mu_j, Sigma_j, shape_j, nu= [0.5, 0.5])
    """
    data = np.asarray(data)
    n, p = data.shape

    kmeans = KMeans(n_clusters=k, random_state=0)
    kmeans.fit(data)

    mu_init = kmeans.cluster_centers_  # shape (k, p)

    # Group points by cluster
    data_cwise = [[] for _ in range(k)]
    for i in range(n):
        data_cwise[kmeans.labels_[i]].append(data[i])
    data_cwise_arr = [np.asarray(arr).T for arr in data_cwise]  # each shape (p, n_j)

    # For shape: sign of third moment (like SkewTMix)
    alpha_init = []
    pi_init = []
    for j in range(k):
        cluster_data = data_cwise_arr[j]  # shape (p, n_j)
        size_j = cluster_data.shape[1]
        pi_init.append(size_j / n)

        if size_j <= 1:
            # degenerate cluster
            Sigma_j = np.eye(p)
            shape_j = np.zeros(p)
        else:
            Sigma_j = np.cov(cluster_data)
            # shape_j from sign( sum( (x - mean)^3 ) ), one approach:
            mean_diff = cluster_data.T - mu_init[j]
            skewness = np.sign(np.sum(mean_diff**3, axis=0))
            shape_j = skewness

        # We'll store (mu_j, Sigma_j, shape_j, nu= [0.5, 0.5]) by default
        alpha_init.append((mu_init[j], Sigma_j, shape_j, np.array([0.5, 0.5])))

    return np.array(pi_init), alpha_init


def estimate_alphas_skewcn(
    X,
    gamma,                   # responsibilities, shape (n, g)
    alpha_prev,              # list of (mu_j, Sigma_j, shape_j, nu_j) 
                             # Here, we assume all clusters share the same nu => the 4th item is the same array for each j
    pi_prev,                 # shape (g,)
    dmvSNC_func, 
    d_mixedmvSNC_func,
    bounds_nu = [(0.01, 0.99), (0.01, 0.99)],
    # You can tune the above bounds if your problem requires different intervals
    **kwargs
):
    """
    Perform one M-step update for a mixture of Skew Contaminated Normal components.
    Follows the logic from the R code (family = "Skew.cn").

    Returns
    -------
    alpha_new : list of length g of (mu_j, Sigma_j, shape_j, nu_new)
    """

    X = np.asarray(X)
    n, p = X.shape
    g = len(alpha_prev)

    # The old shared nu (2D vector)
    nu_old = alpha_prev[0][3]  # shape (2,) e.g. [nu1, nu2]

    # Extract old parameters
    mu_old_list = [alpha_prev[j][0] for j in range(g)]
    Sigma_old_list = [alpha_prev[j][1] for j in range(g)]
    shape_old_list = [alpha_prev[j][2] for j in range(g)]
    # We'll compute Delta_j, Gama_j from shape_j, Sigma_j

    # For numeric stability
    eps = 1e-10

    # 1) Compute d1 = dmvSNC(X, mu_j, Sigma_j, shape_j, nu_old) for each j
    #    mixture pdf d2 = sum_j pi_j * d1_j
    #    tal[i,j] = d1[i,j]* pi_j / d2[i]
    # But we already have gamma=tal from the E-step in the base class. 
    # So we skip recomputing it. 
    # We do, however, need the partial E-step quantities: u[i,j], E[i,j], etc.

    # We define arrays for S1, S2, S3 in each cluster
    S1_mat = np.zeros((n, g))
    S2_mat = np.zeros((n, g))
    S3_mat = np.zeros((n, g))

    # We'll produce "Delta_j_old" and "Gama_j_old" for each cluster 
    # as in the R code:
    Delta_old = []
    Gama_old  = []
    for j in range(g):
        shape_j = shape_old_list[j]
        Sigma_j = Sigma_old_list[j]

        # delta_j = shape_j / sqrt(1 + shape_j' shape_j)
        shape_j = np.asarray(shape_j)
        denom = np.sqrt(1.0 + shape_j @ shape_j)
        delta_j = shape_j / denom

        # Delta_j = sqrtm(Sigma_j) @ delta_j
        sqrt_Sigma_j = sqrtm(Sigma_j + 1e-12*np.eye(p))  # safer
        Delta_j = sqrt_Sigma_j @ delta_j.reshape(-1,1)  # (p,1)
        # Gama_j = Sigma_j - Delta_j Delta_j^T
        Gama_j = Sigma_j - Delta_j @ Delta_j.T

        Delta_old.append(Delta_j)
        Gama_old.append(Gama_j)

    

    for j in range(g):
        mu_j = mu_old_list[j]
        shape_j = shape_old_list[j]
        # We'll compute delta_j, Gama_j, etc from old 
        Delta_j_old = Delta_old[j]
        Gama_j_old  = Gama_old[j]

        # Precompute inv(Gama_j_old)
        inv_Gama_j_old = np.linalg.inv(Gama_j_old + 1e-12*np.eye(p))

        # Compute Mtij2
        tmp_val = (Delta_j_old.T @ inv_Gama_j_old @ Delta_j_old)[0,0]
        Mtij2_j = 1.0/(1.0 + tmp_val)
        Mtij_j  = np.sqrt(Mtij2_j)

        # Compute pdf_j = dmvSNC(X, mu_j, Sigma_j, shape_j, nu_old)
        pdf_j = dmvSNC_func(X, mu_j, Sigma_old_list[j], shape_j, nu_old)
        pdf_j = np.maximum(pdf_j, 1e-300)

        # We'll build A[i], mutij[i]
        Dif = X - mu_j
        mutij_arr = np.zeros(n)
        A_arr = np.zeros(n)

        # factor = Mtij2_j * Delta_j_old^T inv_Gama_j_old
        # shape (1, p) times Dif[i,:] => scalar
        factor_j = (Mtij2_j*(Delta_j_old.T @ inv_Gama_j_old)).ravel()  # shape (p,)

        mutij_arr = np.einsum('j,ij->i', factor_j, Dif)  # shape (n,)
        A_arr = mutij_arr / (Mtij_j + eps)


        nu1, nu2 = nu_old
        dmv_norm1 = dmvnorm(X, mu_j, Sigma_old_list[j]/nu2)
        dmv_norm2 = dmvnorm(X, mu_j, Sigma_old_list[j])

        # For A in the pnorm, see the code: "pnorm( sqrt(nu[2]) * A )" and "pnorm(A)"
        # For E's 'dnorm(...)' the same arguments.
        from scipy.stats import norm

        pnorm_A1 = norm.cdf(np.sqrt(nu2)*A_arr)
        pnorm_A2 = norm.cdf(A_arr)
        dnorm_A1 = norm.pdf(np.sqrt(nu2)*A_arr)
        dnorm_A2 = norm.pdf(A_arr)

        
        u_arr = (2.0/pdf_j)* ( nu1*nu2*dmv_norm1*pnorm_A1 + (1-nu1)*dmv_norm2*pnorm_A2 )
        E_arr = (2.0/pdf_j)* ( nu1*np.sqrt(nu2)*dmv_norm1*dnorm_A1 + (1-nu1)*dmv_norm2*dnorm_A2 )

        # Now S1, S2, S3:
        gamma_j = gamma[:, j]
        S1_j = gamma_j*u_arr
        S2_j = gamma_j*(mutij_arr*u_arr + Mtij_j*E_arr)
        S3_j = gamma_j*(mutij_arr**2*u_arr + Mtij2_j + Mtij_j*mutij_arr*E_arr)

        S1_mat[:, j] = S1_j
        S2_mat[:, j] = S2_j
        S3_mat[:, j] = S3_j

    

    alpha_new_list = []
    for j in range(g):
        # old references
        mu_j_old = mu_old_list[j]
        Delta_j_old = Delta_old[j]
        shape_j_old = shape_old_list[j]
        # sums:
        sum_S1 = np.sum(S1_mat[:, j])
        sum_S2 = np.sum(S2_mat[:, j])
        sum_S3 = np.sum(S3_mat[:, j])

        if sum_S1 < eps:
            # fallback if the cluster is essentially empty
            mu_j_new = mu_j_old
        else:
            # mu_j_new
            sum_S1_X = np.einsum('i,ij->j', S1_mat[:, j], X)
            sum_S2_Delta_old = sum_S2*Delta_j_old.ravel()
            mu_j_new = (sum_S1_X - sum_S2_Delta_old)/(sum_S1 + eps)

        # Then Delta_j_new
        Dif_new = X - mu_j_new
        # sum_i [ S2[i,j]* Dif_new[i] ]
        sum_S2_Dif = np.einsum('i,ij->j', S2_mat[:, j], Dif_new)
        Delta_j_new = sum_S2_Dif/(sum_S3 + eps)
        Delta_j_new = Delta_j_new.reshape(-1,1)

        # Gama_j_new
        sum2 = np.zeros((p,p))
        for i in range(n):
            dif_i = Dif_new[i].reshape(-1,1)
            s1_val = S1_mat[i, j]
            s2_val = S2_mat[i, j]
            s3_val = S3_mat[i, j]
            sum2 += ( s1_val*(dif_i @ dif_i.T)
                       - s2_val*(Delta_j_new @ dif_i.T)
                       - s2_val*(dif_i @ Delta_j_new.T)
                       + s3_val*(Delta_j_new @ Delta_j_new.T) )
        n_j = np.sum(gamma[:, j])
        Gama_j_new = sum2/(n_j + eps)
        Sigma_j_new = Gama_j_new + Delta_j_new @ Delta_j_new.T

        # shape_j_new

        inv_sqrt_Sigma_j_new = np.linalg.inv(sqrtm(Sigma_j_new + 1e-12*np.eye(p)))
        numerator = inv_sqrt_Sigma_j_new @ Delta_j_new
        denom_ = 1.0 - (Delta_j_new.T @ np.linalg.inv(Sigma_j_new + eps*np.eye(p)) @ Delta_j_new)
        denom_ = np.sqrt(np.maximum(1e-300, denom_[0,0]))
        shape_j_new = (numerator/denom_).ravel()

        alpha_new_list.append((mu_j_new, Sigma_j_new, shape_j_new, nu_old))  # temp, we haven't updated nu yet

    # Next: update the shared nu by maximizing sum(log(d_mixedmvSNC(...))).
    # We'll define a function to pass to 'minimize' (L-BFGS-B).
    def neg_loglik_snc(nu_val):
        # nu_val is shape (2,).
        # Build a temp alpha with the same mu, Sigma, shape but new nu
        mu_list  = [a[0] for a in alpha_new_list]
        Sig_list = [a[1] for a in alpha_new_list]
        lam_list = [a[2] for a in alpha_new_list]
        # mixture pdf
        mix_pdf = d_mixedmvSNC_func(X, pi_prev, mu_list, Sig_list, lam_list, nu_val)
        mix_pdf = np.maximum(mix_pdf, 1e-300)
        return -np.sum(np.log(mix_pdf))  # negative for minimization

    # Use scipy.optimize.minimize with L-BFGS-B to update nu
    res = minimize(
        fun=neg_loglik_snc,
        x0=nu_old,  # initial guess
        method='L-BFGS-B',
        bounds=bounds_nu
    )
    nu_new = res.x  # shape(2,)

    # Now finalize alpha_new with the updated nu_new in each cluster
    alpha_final = []
    for j in range(g):
        mu_j, Sig_j, shp_j, _ = alpha_new_list[j]
        alpha_final.append((mu_j, Sig_j, shp_j, nu_new))

    return alpha_final


class SkewContMix(BaseMixture):
    """
    Mixture of Multivariate Skew Contaminated Normal (Skew-CN),
    sharing a single 2D vector nu=[nu1, nu2].
    """

    def __init__(self,
                 n_clusters,
                 tol=1e-4,
                 initialization="kmeans",
                 print_log_likelihood=False,
                 max_iter=25,
                 verbose=True):
        super().__init__(
            n_clusters=n_clusters,
            EM_type="Soft",
            mixture_type="identical",
            tol=tol,
            print_log_likelihood=print_log_likelihood,
            max_iter=max_iter,
            verbose=verbose
        )
        self.k = n_clusters
        self.initialization = initialization
        self.family = "skew_cn"

    def _log_pdf_skewcn(self, X, alphas):
        """
        Evaluate log of Skew-CN pdf for each row in X, for each cluster j.

        alphas[j] = (mu_j, Sigma_j, shape_j, nu_j) 
                    but we assume a single nu_j for all j in practice.
        Returns a (n, g) array of log-densities.
        """
        N, p = X.shape
        g = len(alphas)
        log_pdf = np.empty((N, g))
        for j in range(g):
            mu_j, Sigma_j, shape_j, nu_j = alphas[j]
            dens_j = dmvSNC(X, mu_j, Sigma_j, shape_j, nu_j)
            dens_j = np.maximum(dens_j, 1e-300)
            log_pdf[:, j] = np.log(dens_j)
        return log_pdf

    def _estimate_weighted_log_prob_identical(self, X, alpha, pi):
        """
        Weighted log-prob: log( f_j(X) ) + log(pi_j )
        """
        log_pdf = self._log_pdf_skewcn(X, alpha)   # (n, g)
        log_pi = np.log(pi)                       # (g,)
        return log_pdf + log_pi

    def fit(self, sample):
        start_time = time.time()
        self.data = self._process_data(sample)
        self.n, self.p = self.data.shape
        self.total_parameters = (self.k -1) + self.k* (2*self.p + (self.p*(self.p+1)/2) +2)

        # 1) Initialization
        if self.initialization == "kmeans":
            self.pi_not, self.alpha_not = k_means_init_skewcn(self.data, self.k)
        else:
            raise NotImplementedError("Only 'kmeans' initialization is supported for SkewContMix.")

        self.alpha_temp = self.alpha_not
        self.pi_temp = self.pi_not

        # 2) Provide references for M-step
        #    We'll define small wrappers that link to the above functions.
        def dmvSNC_local(data, mu, Sigma, lam, nu):
            return dmvSNC(data, mu, Sigma, lam, nu)

        def d_mixedmvSNC_local(data, pi_list, mu_list, Sig_list, lam_list, nu):
            return d_mixedmvSNC(data, pi_list, mu_list, Sig_list, lam_list, nu)

        self.dmvSNC_func = dmvSNC_local
        self.d_mixedmvSNC_func = d_mixedmvSNC_local

        # 3) Run the base-class _fit method
        pi_new, alpha_new, log_like_new, log_gamma_new = self._fit(
            self.data,
            self.pi_temp,
            self.alpha_temp,
            estimate_alphas_function=self._estimate_alphas_wrapper
        )

        self.pi_new = pi_new
        self.alpha_new = alpha_new
        self.cluster = log_gamma_new.argmax(axis=1)   # cluster assignment
        self.gamma_temp_ar = np.exp(log_gamma_new)
        self.log_likelihood_new = log_like_new

        if self.verbose:
            print("Mixtures of Skew-CN Fitting Done Successfully")
        end_time = time.time()
        self.execution_time = end_time - start_time

    def _estimate_alphas_wrapper(self, X, gamma, alpha_prev, **kwargs):
        """
        Wrapper that calls estimate_alphas_skewcn for a single M-step.
        """
        pi_prev = self.pi_temp  # from the last E-step
        alpha_new = estimate_alphas_skewcn(
            X,
            gamma,
            alpha_prev,
            pi_prev,
            dmvSNC_func=self.dmvSNC_func,
            d_mixedmvSNC_func=self.d_mixedmvSNC_func
        )
        return alpha_new
    
    def get_params(self):
        #print("The estimated pi values are ", self.pi_new)
        #print("The estimated alpha values are ", self.alpha_new)

        return self.pi_new, self.alpha_new

    def predict(self):
        return self.cluster

    def predict_new(self, x):
        x=self._process_data(x)
        data_lol = x.tolist()
        cluster, _ = mixture_clusters(self.gamma_temp_ar, data_lol)

        return cluster


    def responsibilities(self):

        return self.gamma_temp_ar

    def n_iter(self):
        return len(self.log_likelihoods)

    def get_info_mat(self):
        
        im = info_matrix(self)
        se = standard_errors_from_info_matrix(im)
        
        return im, se


from fmvmm.mixtures import skewlaplacemix
import numpy as np
from fmvmm.mixtures.skewlaplacemix import compute_empirical_info_and_se


def logpdf(x,mu,sigma,gamma):
    
    return skewlaplacemix.compute_log_pdf_skewlaplace(x, mu, sigma, gamma)

def pdf(x,mu,sigma,gamma):
    
    return np.exp(logpdf(x,mu,sigma,gamma))
    
def loglike(x,mu,sigma,gamma):
    
    np.sum(logpdf(x,mu,sigma,gamma))
    
def total_params(mu,sigma,gamma):
    p = len(mu)
    
    return 2*p + (p*(p+1)/2)

def rvs(mu, sigma, gamma, size=1):
    
    return skewlaplacemix.sample_skewlaplace(mu, sigma, gamma, size)

def mean(mu, sigma, gamma):
    """
    mu:    shape (p,)
    sigma: shape (p, p) [not actually used here]
    gamma: shape (p,)

    Returns the mean vector of dimension (p,).
    """
    p = len(mu)
    # mu + (p+1)*gamma => shape (p,)
    return mu + (p + 1) * gamma

def var(mu, sigma, gamma):
    """
    mu:    shape (p,)
    sigma: shape (p, p)
    gamma: shape (p,)

    Returns the 'variance' or second-moment matrix of dimension (p, p).
    Often used in GH-like families: (p+1)*(sigma + gamma gamma^T).
    """
    p = len(mu)
    # Outer product of gamma with itself => shape (p,p)
    gamma_outer = np.outer(gamma, gamma)

    # (p+1)*(sigma + gamma gamma^T)
    return (p + 1) * (sigma + gamma_outer)


def fit(x):
    model = skewlaplacemix.SkewLaplaceMix(1, verbose = False)
    model.fit(x)
    _, alphas = model.get_params()
    
    return alphas[0][0], alphas[0][1], alphas[0][2]

def info_mat(X,mu,sigma,lmbda):
    
    p =len(mu)
    g = 1
    n = X.shape[0]
    
    IM, _ = compute_empirical_info_and_se(None, pi = [1],
            alpha= [[mu, sigma, lmbda]],data=X,gamma_res=np.array([np.ones(n)]).T, use_model=False)
    
    final_IM = expand_reduced_IM_to_full_no_pi(IM, p, g)
    
    return final_IM

def expand_reduced_IM_to_full_no_pi(IM_reduced, p, g):
    """
    Expand and remove mixture-weight parameters from the 'reduced' info matrix
    of a Skew-Laplace mixture, as produced by info_matrix_skewnormal(...).

    In info_matrix_skewnormal, each cluster j has:
      - mu_j : p
      - shape_j : p
      - Sigma_j : p*(p+1)//2   (upper-triangular portion)
    => cluster_params_reduced = 2*p + (p*(p+1)//2)
    => total dimension:  g*cluster_params_reduced + (g-1 if g>1 else 0)

    Steps:
      1) Remove (g-1) rows/cols for pi_j if g>1.
      2) Expand each Sigma_j from p*(p+1)//2 to p^2.
      3) Return the expanded info matrix with dimension
         g*(2p + p^2), i.e. only mu_j, shape_j, full Sigma_j.

    Parameters
    ----------
    IM_reduced : ndarray of shape (M_reduced, M_reduced)
        The reduced info matrix from info_matrix_skewnormal(...).
    p : int
        Dimension of each component (size of mu_j, shape_j).
    g : int
        Number of mixture components.

    Returns
    -------
    IM_full_no_pi : ndarray of shape (M_full_no_pi, M_full_no_pi)
        The expanded info matrix, with no pi_j parameters and
        full (p^2) blocks for each Sigma_j.
    """
    # 1) Dimensions in the "reduced" format
    cluster_params_reduced = 2*p + (p*(p+1)//2)
    M_reduced_total = g*cluster_params_reduced + (g-1 if g>1 else 0)  # no extra 'nu' in pure Skew-Normal
    if IM_reduced.shape != (M_reduced_total, M_reduced_total):
        raise ValueError(
            f"Expected IM_reduced to be {M_reduced_total} x {M_reduced_total}, "
            f"but got {IM_reduced.shape}."
        )

    # 2) After removing pi block (if g>1), we have:
    M_reduced_no_pi = g*cluster_params_reduced

    # 3) After expanding Sigma to p^2 for each cluster, final dimension is:
    #    M_full_no_pi = g*(2p + p^2).
    M_full_no_pi = g * (2*p + p*p)

    # --- Remove the pi block if g>1
    if g > 1:
        offset_pi = g*cluster_params_reduced  # start of pi block
        keep_mask = np.ones(M_reduced_total, dtype=bool)
        # The pi parameters are offset_pi : offset_pi + (g-1)
        pi_indices = np.arange(offset_pi, offset_pi + (g-1))
        keep_mask[pi_indices] = False
        IM_reduced_no_pi = IM_reduced[keep_mask][:, keep_mask]
    else:
        # If g=1, no pi
        IM_reduced_no_pi = IM_reduced

    if IM_reduced_no_pi.shape != (M_reduced_no_pi, M_reduced_no_pi):
        raise ValueError(
            f"After removing pi, expected shape ({M_reduced_no_pi}, {M_reduced_no_pi}), "
            f"but got {IM_reduced_no_pi.shape}."
        )

    # --- 4) Build duplication matrix D_no_pi
    # shape: (M_full_no_pi, M_reduced_no_pi)
    D_no_pi = np.zeros((M_full_no_pi, M_reduced_no_pi), dtype=float)

    offset_reduced = 0
    offset_full = 0

    def copy_block_1to1(num_params):
        # copies 'num_params' parameters in a 1-to-1 fashion
        nonlocal offset_reduced, offset_full
        for kk in range(num_params):
            D_no_pi[offset_full + kk, offset_reduced + kk] = 1.0
        offset_reduced += num_params
        offset_full    += num_params

    # For each cluster j => we have mu_j (p), shape_j (p), Sigma_j (p*(p+1)//2)
    for j_ in range(g):
        # mu_j => p
        copy_block_1to1(p)
        # shape_j => p
        copy_block_1to1(p)

        # Expand Sigma_j from p*(p+1)//2 => p^2
        old_start = offset_reduced
        new_start = offset_full
        old_index = 0

        for row_ in range(p):
            for col_ in range(row_, p):
                old_idx = old_start + old_index
                old_index += 1
                new_idx1 = new_start + (row_*p + col_)
                new_idx2 = new_start + (col_*p + row_)

                D_no_pi[new_idx1, old_idx] = 1.0
                if new_idx2 != new_idx1:
                    D_no_pi[new_idx2, old_idx] = 1.0

        offset_reduced += (p*(p+1)//2)
        offset_full    += (p*p)

    # --- 5) Final expanded info matrix
    IM_full_no_pi = D_no_pi @ IM_reduced_no_pi @ D_no_pi.T
    return IM_full_no_pi



import numpy as np
from scipy.special import gammaln,logsumexp
import sys

import numpy as np
from numpy import (
    asanyarray,
    exp
)
from numpy.linalg import norm
from scipy.special import polygamma, psi
from scipy.special import digamma
import pandas as pd
import copy
from fmvmm.utils.utils_dmm import (kmeans_init, gmm_init, kmeans_init_adv, gmm_init_adv, random_init,
                             dmm_loglikelihood, dmm_responsibilities, dmm_pi_estimate)

from fmvmm.utils.utils_mixture import (mixture_clusters)
import time
from multiprocessing import Pool
from fmvmm.mixtures._base import BaseMixture
from fmvmm.utils.utils_dmm import combined_info_and_se

MAXINT = sys.maxsize
euler = -1 * psi(1)  # Euler-Mascheroni constant



def get_precision_es(data):
    # Convert the list of lists to a NumPy array
    data_array = np.array(data)
    # print(data_array.shape)

    # Calculate mean (mu_1) and variance (var_1) for the first column (axis 0)
    mu_1 = np.mean(data_array[:, 0])
    var_1 = np.var(data_array[:, 0])

    # Compute the expression
    result = (mu_1 - (var_1 + mu_1**2)) / var_1

    return result

def get_means(data):
    # Convert the list of lists to a NumPy array
    data_array = np.array(data)
    # print(data_array.shape)

    # Calculate mean
    mu = [np.mean(data_array[:, i]) for i in range(data_array.shape[1])]


    return mu


class NotConvergingError(Exception):
    """Error when a successive approximation method doesn't converge
    """
    pass


def _ipsi(y, tol=1.48e-9, maxiter=10):
    """Inverse of psi (digamma) using Newton's method. For the purposes
    of Dirichlet MLE, since the parameters a[i] must always
    satisfy a > 0, we define ipsi :: R -> (0,inf).

    Parameters
    ----------
    y : (K,) shape array
        y-values of psi(x)
    tol : float
        If Euclidean distance between successive parameter arrays is less than
        ``tol``, calculation is taken to have converged.
    maxiter : int
        Maximum number of iterations to take calculations. Default is 10.

    Returns
    -------
    (K,) shape array
        Approximate x for psi(x)."""
    y = asanyarray(y, dtype="float")
    x0 = np.piecewise(
        y,
        [y >= -2.22, y < -2.22],
        [(lambda x: exp(x) + 0.5), (lambda x: -1 / (x + euler))],
    )
    for i in range(maxiter):
        x1 = x0 - (psi(x0) - y) / _trigamma(x0)
        if norm(x1 - x0) < tol:
            return x1
        x0 = x1
    # raise NotConvergingError(
    #     f"Failed to converge after {maxiter} iterations, " f"value is {x1}")
    return x1


def _trigamma(x):
    return polygamma(1, x)


def dirichlet_mix_mle(x: np.ndarray, gamma: np.ndarray, alpha_init):

    p = x.shape[1]
    n = x.shape[0]
    k = gamma.shape[1]
    x_lol = x.tolist()
    alpha_not = alpha_init
    N_js = [np.sum(gamma[:, l]) for l in range(k)]
    alpha_old = alpha_not
    alpha_all = []
    for o in range(k):

        log_x = []
        for f in range(p):
            log_x_temp = (
                1/N_js[o])*np.sum([np.log(x_lol[j][f])*gamma[j, o] for j in range(n)])
            log_x.append(log_x_temp)
        alpha_new = []
        for i in range(p):
            diff = 5
            t = copy.deepcopy(alpha_old[o][i])
            t1 = copy.deepcopy(alpha_old[o][i])
            di_alpha_not = digamma(t)
            di_alpha_old = di_alpha_not
            beta = np.sum(alpha_old[o])-t
            while diff > 0.0001:
                di_alpha_sum = digamma(np.sum(alpha_old[o])-t + t1)
                di_alpha_new = di_alpha_sum + log_x[i]
                t1 = _ipsi(di_alpha_new, tol=1.48e-9, maxiter=1000)
                diff = abs(di_alpha_new - di_alpha_old)
                di_alpha_old = di_alpha_new
                alpha_new_temp = t1
            alpha_new.append(alpha_new_temp)
        alpha_all.append(alpha_new)

    return alpha_all

def parallel_alpha_computation(args):
    t, alpha_old_i, log_x_i = args
    beta = np.sum(alpha_old_i) - t
    alpha_new_temp = _ipsi(np.log(beta) + log_x_i, tol=1.48e-9, maxiter=10000)
    return alpha_new_temp




def dirichlet_mix_mle_highdimensional(x: np.ndarray, gamma: np.ndarray, alpha_init):

    p = x.shape[1]
    n = x.shape[0]
    k = gamma.shape[1]
    alpha_not = alpha_init
    N_js = np.sum(gamma, axis=0)
    log_x = (1 / N_js) * np.dot(np.log(x).T, gamma)
    alpha_old = alpha_not
    alpha_all = []
    for o in range(k):
        alpha_new = []
        with Pool() as pool:
            # Prepare arguments for parallel computation
            args_list = [(copy.deepcopy(alpha_old[o][i]), copy.deepcopy(alpha_old[o]), log_x[i,o]) for i in range(p)]
            
            # Use the Pool to parallelize the computation
            alpha_new = pool.map(parallel_alpha_computation, args_list)
        # for i in range(p):
        #     t = copy.deepcopy(alpha_old[o][i])
        #     beta = np.sum(alpha_old[o])-t
        #     alpha_new_temp = _ipsi(
        #         np.log(beta)+log_x[i,o], tol=1.48e-9, maxiter=10000)
        #     alpha_new.append(alpha_new_temp)
        alpha_all.append(alpha_new)

    return alpha_all


def parallel_alpha_computation2(args):
    t,beta, log_x_i = args
    beta=beta-t
    alpha_new_temp = _ipsi(np.log(beta) + log_x_i, tol=1.48e-9, maxiter=10000)
    return alpha_new_temp





def dirichlet_mix_mle_approx(x: np.ndarray, gamma: np.ndarray, alpha_init):

    p = x.shape[1]
    n = x.shape[0]
    k = gamma.shape[1]
    alpha_not = alpha_init
    N_js = np.sum(gamma, axis=0)
    log_x = (1 / N_js) * np.dot(np.log(x).T, gamma)
    # x_lol = x.tolist()
    # alpha_not = alpha_init
    # N_js = [np.sum(gamma[:, l]) for l in range(k)]
    alpha_old = alpha_not
    alpha_all = []
    _,data_cwise=mixture_clusters(gamma, x.tolist())
    for o in range(k):

        if np.array(data_cwise[o]).ndim == 1:
            alpha_new = alpha_init[o]
        else:
            beta=get_precision_es(data_cwise[o])
            alpha_new = []
            with Pool(processes=16) as pool:
                # Prepare arguments for parallel computation
                args_list = [(copy.deepcopy(alpha_old[o][i]), beta, log_x[i,o]) for i in range(p)]
                
                # Use the Pool to parallelize the computation
                alpha_new = pool.map(parallel_alpha_computation2, args_list)
            
        alpha_all.append(alpha_new)

    return alpha_all







def dirichlet_mean_precision_mle(x: np.ndarray, gamma: np.ndarray, alpha_init):
    alpha_old = alpha_init
    max_iter=100
    diff = 5
    norm0 = 5
    it=0
    while diff > 0.0001:
        alpha_1 = _fit_s(x, gamma, alpha_old)
        alpha_2 = _fit_m(x, gamma, alpha_1)
        norm1 = norm(np.array(alpha_2)-np.array(alpha_old))
        diff = abs(norm1-norm0)
        #print("alpha_diff",diff)
        norm0 = norm1
        alpha_old = alpha_2
    #     it=it+1
    #     if it==max_iter:
    #         return alpha_2
    # print(it)
    return alpha_2


def dirichlet_mean_identical_precision_mle(x: np.ndarray, gamma: np.ndarray, alpha_init):
    alpha_old = alpha_init
    diff = 5
    norm0 = 5
    while diff > 0.0001:
        alpha_1 = _fit_s_identical(x, gamma, alpha_old)
        alpha_2 = _fit_m(x, gamma, alpha_1)
        norm1 = norm(np.array(alpha_2)-np.array(alpha_old))
        diff = abs(norm1-norm0)
        #print("alpha_diff",diff)
        norm0 = norm1
        alpha_old = alpha_2
    return alpha_2

def dirichlet_mean_precision_mle_approx(x: np.ndarray, gamma: np.ndarray, alpha_init):
    alpha_old = alpha_init
    diff = 5
    norm0 = 5
    while diff > 0.0001:
        alpha_1 = _fit_s_approx(x, gamma, alpha_old)
        alpha_2 = _fit_m(x, gamma, alpha_1)
        norm1 = norm(np.array(alpha_2)-np.array(alpha_old))
        diff = abs(norm1-norm0)
        #print("alpha_diff",diff)
        norm0 = norm1
        alpha_old = alpha_2
    return alpha_2

def dirichlet_mean_identical_precision_mle_approx(x: np.ndarray, gamma: np.ndarray, alpha_init):
    alpha_old = alpha_init
    diff = 5
    norm0 = 5
    while diff > 0.0001:
        alpha_1 = _fit_s_ideantical_approx(x, gamma, alpha_old)
        alpha_2 = _fit_m(x, gamma, alpha_1)
        norm1 = norm(np.array(alpha_2)-np.array(alpha_old))
        diff = abs(norm1-norm0)
        #print("alpha_diff",diff)
        norm0 = norm1
        alpha_old = alpha_2
    return alpha_2



def dirichlet_mean_known_precision_mle(x: np.ndarray, gamma: np.ndarray, alpha_init,true_s):
    
    alpha_1 = _fit_m_known_s(x, gamma, alpha_init,true_s)
    return alpha_1


def dirichlet_known_mean_precision_mle(x: np.ndarray, gamma: np.ndarray, alpha_init,true_m):

    alpha_1 = _fit_s_known_m(x, gamma, alpha_init,true_m)
    return alpha_1



def _fit_s(x: np.ndarray, gamma: np.ndarray, alpha_old):

    p = x.shape[1]
    n = x.shape[0]
    k = gamma.shape[1]
    x_lol = x.tolist()
    N_js = [np.sum(gamma[:, l]) for l in range(k)]
    sj_old = [np.sum(alpha_old[i]) for i in range(len(alpha_old))]
    mj_old = [(np.array(alpha_old[i])/sj_old[i]).tolist()
              for i in range(len(alpha_old))]
    alpha_new_all = []
    for o in range(k):
        diff = 5
        sjold = sj_old[o]
        max_iter=100
        it=0
        while diff > 0.0001:
            cs = []
            for i in range(n):
                for j in range(p):
                    c_temp = mj_old[o][j]*gamma[i, o]*np.log(x_lol[i][j])
                    cs.append(c_temp)
            c = np.sum(cs)
            bs = []
            for i in range(p):
                b_temp = mj_old[o][i]*digamma(sjold*mj_old[o][i])
                bs.append(b_temp)
            b = N_js[o]*np.sum(bs)
            a = N_js[o]*digamma(sjold)
            firststdif = a-b+c
            d = N_js[o]*_trigamma(sjold)
            es = []
            for i in range(p):
                es_temp = (mj_old[o][i]**2)*_trigamma(sjold*mj_old[o][i])
                es.append(es_temp)
            e = N_js[o]*np.sum(es)
            seconddif = d-e
            # if seconddif < 0:
            #     sj_new = sjold - firststdif/seconddif
            # elif firststdif + sjold*seconddif < 0:
            #     sj_new = 1/((1/sjold)+(1/sjold**2)*(1/seconddif)*firststdif)
            # else:
            #     raise NotConvergingError(f"Unable to update s from {sjold}")
            if firststdif + sjold*seconddif < 0:
                sj_new = 1/((1/sjold)+(1/sjold**2)*(1/seconddif)*firststdif)
            elif seconddif < 0:
                sj_new = sjold - firststdif/seconddif
            else:
                raise NotConvergingError(f"Unable to update s from {sjold}")
            diff = abs(sj_new-sjold)
            #print("s_diff",diff)
            sjold = sj_new
            it=it+1
            if it==max_iter:
                break
        # print(it)
        alphaj_new = [sj_new*mj_old[o][i] for i in range(p)]
        alpha_new_all.append(alphaj_new)
    return alpha_new_all


def _fit_m(x: np.ndarray, gamma: np.ndarray, alpha_old):
    p = x.shape[1]
    n = x.shape[0]
    k = gamma.shape[1]
    x_lol = x.tolist()
    N_js = [np.sum(gamma[:, l]) for l in range(k)]
    sj_old = [np.sum(alpha_old[i]) for i in range(len(alpha_old))]
    mj_old = [(np.array(alpha_old[i])/sj_old[i]).tolist()
              for i in range(len(alpha_old))]
    alpha_new_all = []

    for o in range(k):

        sjold = sj_old[o]
        mjold = mj_old[o]
        alpha_j_old = alpha_old[o]
        m6 = []
        for s in range(p):
            m5 = np.sum([gamma[i, o]*np.log(x_lol[i][s])
                        for i in range(n)])/N_js[o]
            m6.append(m5)
        diff = 5
        norm0 = 5
        while diff > 0.0001:
            alpha_j_new = []
            for j in range(p):
                m1 = m6[j]
                m2 = [mjold[q]*m6[q]for q in range(p)]
                m3 = np.sum(m2)
                m4 = np.sum([mjold[r]*digamma(sjold*mjold[r])
                            for r in range(p)])
                alpha_jm_new = _ipsi(m1-m3+m4, tol=1.48e-9, maxiter=1000)
                alpha_j_new.append(alpha_jm_new)
            alpha_j_new = np.array(alpha_j_new) / \
                np.array(alpha_j_new).sum() * sjold
            norm1 = norm(alpha_j_new-alpha_j_old)
            diff = abs(norm1-norm0)
            #print("m_diff",diff)
            norm0 = norm1
            mjold = alpha_j_new/alpha_j_new.sum()
        alpha_new_all.append(alpha_j_new.tolist())
    return alpha_new_all


def _fit_m_known_s(x: np.ndarray, gamma: np.ndarray, alpha_old,true_s):
    p = x.shape[1]
    n = x.shape[0]
    k = gamma.shape[1]
    x_lol = x.tolist()
    N_js = [np.sum(gamma[:, l]) for l in range(k)]
    _,data_cwise=mixture_clusters(gamma.tolist(), x.tolist())
    if true_s==None:
        sj_old=[get_precision_es(data_cwise[o]) for o in range(k)]
    else:
        sj_old=true_s
    # sj_old = [np.sum(alpha_old[i]) for i in range(len(alpha_old))]
    mj_old = [(np.array(alpha_old[i])/sj_old[i]).tolist()
              for i in range(len(alpha_old))]
    alpha_new_all = []

    for o in range(k):

        sjold = sj_old[o]
        mjold = mj_old[o]
        alpha_j_old = alpha_old[o]
        m6 = []
        for s in range(p):
            m5 = np.sum([gamma[i, o]*np.log(x_lol[i][s])
                        for i in range(n)])/N_js[o]
            m6.append(m5)
        diff = 5
        norm0 = 5
        while diff > 0.0001:
            alpha_j_new = []
            for j in range(p):
                m1 = m6[j]
                m2 = [mjold[q]*m6[q]for q in range(p)]
                m3 = np.sum(m2)
                m4 = np.sum([mjold[r]*digamma(sjold*mjold[r])
                            for r in range(p)])
                alpha_jm_new = _ipsi(m1-m3+m4, tol=1.48e-9, maxiter=1000)
                alpha_j_new.append(alpha_jm_new)
            alpha_j_new = np.array(alpha_j_new) / \
                np.array(alpha_j_new).sum() * sjold
            norm1 = norm(alpha_j_new-alpha_j_old)
            diff = abs(norm1-norm0)
            #print("m_diff",diff)
            norm0 = norm1
            mjold = alpha_j_new/alpha_j_new.sum()
        alpha_new_all.append(alpha_j_new.tolist())
    return alpha_new_all





def _fit_s_identical(x: np.ndarray, gamma: np.ndarray, alpha_old):

    p = x.shape[1]
    n = x.shape[0]
    k = gamma.shape[1]
    x_lol = x.tolist()
    N_js = [np.sum(gamma[:, l]) for l in range(k)]
    sj_old = [np.sum(alpha_old[i]) for i in range(len(alpha_old))]
    mj_old = [(np.array(alpha_old[i])/sj_old[i]).tolist()
              for i in range(len(alpha_old))]
    alpha_new_all = []
    diff = 5
    sjold = sj_old[0]
    while diff > 0.0001:
        cs = []
        for i in range(n):
            for j in range(p):
                c_temp = mj_old[0][j]*gamma[i, 0]*np.log(x_lol[i][j])
                cs.append(c_temp)
        c = np.sum(cs)
        bs = []
        for i in range(p):
            b_temp = mj_old[0][i]*digamma(sjold*mj_old[0][i])
            bs.append(b_temp)
        b = N_js[0]*np.sum(bs)
        a = N_js[0]*digamma(sjold)
        firststdif = a-b+c
        d = N_js[0]*_trigamma(sjold)
        es = []
        for i in range(p):
            es_temp = (mj_old[0][i]**2)*_trigamma(sjold*mj_old[0][i])
            es.append(es_temp)
        e = N_js[0]*np.sum(es)
        seconddif = d-e
        # if seconddif < 0:
        #     sj_new = sjold - firststdif/seconddif
        # elif firststdif + sjold*seconddif < 0:
        #     sj_new = 1/((1/sjold)+(1/sjold**2)*(1/seconddif)*firststdif)
        if firststdif + sjold*seconddif < 0:
            sj_new = 1/((1/sjold)+(1/sjold**2)*(1/seconddif)*firststdif)
        elif seconddif < 0:
            sj_new = sjold - firststdif/seconddif
        else:
            raise NotConvergingError(f"Unable to update s from {sjold}")
        diff = abs(sj_new-sjold)
        #print("s_diff",diff)
        sjold = sj_new
    for o in range(k):
        alphaj_new = [sj_new*mj_old[o][i] for i in range(p)]
        alpha_new_all.append(alphaj_new)
    return alpha_new_all


def _fit_s_approx(x: np.ndarray, gamma: np.ndarray, alpha_old):

    p = x.shape[1]
    n = x.shape[0]
    k = gamma.shape[1]
    x_lol = x.tolist()
    N_js = [np.sum(gamma[:, l]) for l in range(k)]
    sj_old = [np.sum(alpha_old[i]) for i in range(len(alpha_old))]
    mj_old = [(np.array(alpha_old[i])/sj_old[i]).tolist()
              for i in range(len(alpha_old))]
    alpha_new_all = []
    _,data_cwise=mixture_clusters(gamma.tolist(), x.tolist())
    for o in range(k):
        beta=get_precision_es(data_cwise[o])
        # sjold = sj_old[o]
        cs = []
        for i in range(n):
            for j in range(p):
                c_temp = mj_old[o][j]*gamma[i, o]*np.log(x_lol[i][j])
                cs.append(c_temp)
        c = np.sum(cs)/N_js[o]
        bs = []
        for i in range(p):
            b_temp = mj_old[o][i]*digamma(beta*mj_old[o][i])
            bs.append(b_temp)
        b = np.sum(bs)
        sj_new=_ipsi(
            b-c, tol=1.48e-9, maxiter=10000)
        alphaj_new = [sj_new*mj_old[o][i] for i in range(p)]
        alpha_new_all.append(alphaj_new)
    return alpha_new_all

def _fit_s_ideantical_approx(x: np.ndarray, gamma: np.ndarray, alpha_old):

    p = x.shape[1]
    n = x.shape[0]
    k = gamma.shape[1]
    x_lol = x.tolist()
    N_js = [np.sum(gamma[:, l]) for l in range(k)]
    sj_old = [np.sum(alpha_old[i]) for i in range(len(alpha_old))]
    mj_old = [(np.array(alpha_old[i])/sj_old[i]).tolist()
              for i in range(len(alpha_old))]
    alpha_new_all = []
    _,data_cwise=mixture_clusters(gamma.tolist(), x.tolist())
    beta=get_precision_es(data_cwise[0])
    # sjold = sj_old[o]
    cs = []
    for i in range(n):
        for j in range(p):
            c_temp = mj_old[0][j]*gamma[i, 0]*np.log(x_lol[i][j])
            cs.append(c_temp)
    c = np.sum(cs)/N_js[0]
    bs = []
    for i in range(p):
        b_temp = mj_old[0][i]*digamma(beta*mj_old[0][i])
        bs.append(b_temp)
    b = np.sum(bs)
    sj_new=_ipsi(
        b-c, tol=1.48e-9, maxiter=10000)
    for o in range(k):
        alphaj_new = [sj_new*mj_old[o][i] for i in range(p)]
        alpha_new_all.append(alphaj_new)
    return alpha_new_all


def _fit_s_known_m(x: np.ndarray, gamma: np.ndarray, alpha_old,true_m):

    p = x.shape[1]
    n = x.shape[0]
    k = gamma.shape[1]
    x_lol = x.tolist()
    N_js = [np.sum(gamma[:, l]) for l in range(k)]
    sj_old = [np.sum(alpha_old[i]) for i in range(len(alpha_old))]
    # sj_old = [28, 92, 60, 4.8]
    # mj_old = [(np.array(alpha_old[i])/sj_old[i]).tolist()
    #           for i in range(len(alpha_old))]
    # print(mj_old)
    _,data_cwise=mixture_clusters(gamma.tolist(), x.tolist())
    if true_m ==None:
        mj_old=[get_means(data_cwise[i]) for i in range(len(alpha_old))]
    else:
        mj_old=true_m
    # print(mj_old)
    alpha_new_all = []
    for o in range(k):
        diff = 5
        sjold = sj_old[o]
        while diff > 0.0001:
            cs = []
            for i in range(n):
                for j in range(p):
                    c_temp = mj_old[o][j]*gamma[i, o]*np.log(x_lol[i][j])
                    cs.append(c_temp)
            c = np.sum(cs)
            bs = []
            for i in range(p):
                b_temp = mj_old[o][i]*digamma(sjold*mj_old[o][i])
                bs.append(b_temp)
            b = N_js[o]*np.sum(bs)
            a = N_js[o]*digamma(sjold)
            firststdif = a-b+c
            d = N_js[o]*_trigamma(sjold)
            es = []
            for i in range(p):
                es_temp = (mj_old[o][i]**2)*_trigamma(sjold*mj_old[o][i])
                es.append(es_temp)
            e = N_js[o]*np.sum(es)
            seconddif = d-e
            # if seconddif < 0:
            #     sj_new = sjold - firststdif/seconddif
            # elif firststdif + sjold*seconddif < 0:
            #     sj_new = 1/((1/sjold)+(1/sjold**2)*(1/seconddif)*firststdif)
            if firststdif + sjold*seconddif < 0:
                sj_new = 1/((1/sjold)+(1/sjold**2)*(1/seconddif)*firststdif)
            elif seconddif < 0:
                sj_new = sjold - firststdif/seconddif
            else:
                raise NotConvergingError(f"Unable to update s from {sjold}")
            diff = abs(sj_new-sjold)
            #print("s_diff",diff)
            sjold = sj_new
        alphaj_new = [sj_new*mj_old[o][i] for i in range(p)]
        alpha_new_all.append(alphaj_new)
    return alpha_new_all




def estimate_alphas(data: np.ndarray, gamma_temp_ar, alpha_temp, method,true_m,true_s):
    if method == "fixediteration":
        alpha_new = dirichlet_mix_mle(data, gamma_temp_ar, alpha_temp)
    elif method == "meanprecision":
        alpha_new = dirichlet_mean_precision_mle(
            data, gamma_temp_ar, alpha_temp)
    # elif method == "highdimensional":
    #     alpha_new = dirichlet_mix_mle_highdimensional(data, gamma_temp_ar, alpha_temp)
    elif method == "highdimensional":
        alpha_new = dirichlet_mix_mle_approx(data, gamma_temp_ar, alpha_temp)
    elif method == "meanprecision_identical":
        alpha_new = dirichlet_mean_identical_precision_mle(data, gamma_temp_ar, alpha_temp)
    elif method == "meanprecision_approx":
        alpha_new = dirichlet_mean_precision_mle_approx(
            data, gamma_temp_ar, alpha_temp)
    elif method == "meanprecision_identical_approx":
        alpha_new = dirichlet_mean_identical_precision_mle_approx(data, gamma_temp_ar, alpha_temp)
    elif method == "meanprecision_known_mean":
        alpha_new = dirichlet_known_mean_precision_mle(data, gamma_temp_ar, alpha_temp,true_m)
    elif method == "meanprecision_known_precision":
        alpha_new = dirichlet_mean_known_precision_mle(data, gamma_temp_ar, alpha_temp,true_s)

    return alpha_new



class DMM_Soft(BaseMixture):
    def __init__(self,n_clusters,tol=0.0001,initialization="kmeans",method="meanprecision",print_log_likelihood=False,max_iter=25, verbose=True):
        super().__init__(n_clusters=n_clusters, EM_type="Soft", mixture_type="identical", tol=tol, print_log_likelihood=print_log_likelihood, max_iter=max_iter, verbose=verbose)
        self.k=n_clusters
        self.initialization=initialization
        self.method = method
        self.family = "dirichlet"
        
    def _log_pdf_dirichlet(self,X,alphas):
        threshold = 1e-10
        N,p=X.shape
        k=len(alphas)
        probs=np.empty((N, k))
        for j in range(k):
            alpha=alphas[j]
            for i in range(N):
                with np.errstate(under="ignore"):
                    x=X[i, :]
                    t1=[(alphm-1)*np.log(xm) for alphm,xm in zip(alpha,x)]
                    # t1 = [(alphm - 1) * (np.log1p(xm - 1) if xm < threshold else np.log(xm)) for alphm, xm in zip(alpha, x)]
                    t2=np.sum(gammaln(alpha))
                    t3=gammaln(np.sum(alpha))
                    probs[i,j]=np.sum(t1)-t2+t3
        return probs    
    
    def _estimate_weighted_log_prob_identical(self, X, alpha, pi):
        return self._log_pdf_dirichlet(X,alpha) + np.log(pi)
    
    def fit(self,sample,true_m=None,true_s=None):
        start_time = time.time()
        self.data = self._process_data(sample)
        self.n, self.p = self.data.shape
        self.total_parameters = (self.k -1) + self.k* (self.p)
        np.random.seed(0)
        # if self.initialization == "kmeans":
        #     self.pi_not, self.alpha_not = kmeans_init(self.data, self.k)
        # elif self.initialization == "gmm":
        #     self.pi_not, self.alpha_not = gmm_init(self.data, self.k)
        if self.initialization == "kmeans":
            self.pi_not, self.alpha_not = kmeans_init_adv(self.data, self.k)
        elif self.initialization == "gmm":
            self.pi_not, self.alpha_not = gmm_init_adv(self.data, self.k)
        elif self.initialization == "random":
            self.pi_not, self.alpha_not = random_init(self.data, self.k)

        self.alpha_temp = self.alpha_not
        self.pi_temp = self.pi_not
        pi_new,alpha_new, log_likelihood_new,log_gamma_new=self._fit(self.data,self.pi_temp,self.alpha_temp,estimate_alphas,method=self.method,true_m=true_m,true_s=true_s)
        estimated_mean = []
        for a in alpha_new:
            mean_temp = [b/np.sum(a) for b in a]
            estimated_mean.append(mean_temp)
        self.pi_new = pi_new
        self.alpha_new = alpha_new
        self.estimated_mean = estimated_mean
        self.cluster = log_gamma_new.argmax(axis=1)
        # self.data_cwise = data_cwise
        self.gamma_temp_ar = np.exp(log_gamma_new)
        # self.gamma_matrix = gamma_matrix
        self.log_likelihood_new = log_likelihood_new
        if self.verbose:
            print("Soft DMM Fitting Done Successfully")
        end_time = time.time()
        self.execution_time=end_time-start_time
    
    
    def get_params(self):
        #print("The estimated pi values are ", self.pi_new)
        #print("The estimated alpha values are ", self.alpha_new)

        return self.pi_new, self.alpha_new

    def predict(self):
        return self.cluster

    def predict_new(self, x):
        x=self._process_data(x)
        data_lol = x.tolist()
        cluster, _ = mixture_clusters(self.gamma_temp_ar, data_lol)

        return cluster

    def get_mean(self):

        return self.estimated_mean

    def get_precision(self):
        preci=[np.sum(al) for al in self.alpha_new]
        
        return preci

    def responsibilities(self):

        return self.gamma_temp_ar

    def n_iter(self):
        return len(self.log_likelihoods)
    
    def get_info_mat(self):
        IM, SE = combined_info_and_se(self.pi_new, np.array(self.alpha_new), self.gamma_temp_ar, mode = "soft")
        
        return IM, SE
        
        
        
        

    # def clustered_data(self):
    #     return pd.DataFrame(self.data_cwise).transpose()

    # def bic(self):

    #     return ((self.k-1)+(self.k*self.p))*np.log(self.n) - 2*(self.log_likelihood_new)

    # def aic(self):

    #     return 2*((self.k-1)+(self.k*self.p)) - 2*(self.log_likelihood_new)

    # def icl(self):
    #     entropy_s=[]
    #     for i in self.gamma_matrix:
    #         for j in i:
    #                 ent_temp=j*np.log1p(j)
    #                 entropy_s.append(ent_temp)



    #     entropy=np.sum(entropy_s)

    #     return ((self.k-1)+(self.k*self.p))*np.log(self.n) - 2*(self.log_likelihood_new) - entropy

    


"""
some matlab codes were received personally from Prof.Fatma Zehra Doğru,
author of Finite Mixtures of Multivariate Skew Laplace Distributions.
Many thanks to her. Rest was incorporated from the paper itself.

"""

import numpy as np
from fmvmm.mixtures._base import BaseMixture
from sklearn.cluster import KMeans
from sklearn.mixture import GaussianMixture
import time
from fmvmm.utils.utils_mixture import (mixture_clusters)
from numpy.linalg import cholesky
import math
from scipy.special import gamma as gamma_func

###############################################################################
# (A) Generate random samples from a SINGLE Multivariate Skew-Laplace
###############################################################################

def sample_skewlaplace(mu, Sigma, gamma,n, random_state=None):
    """
    Generate 'n' samples in R^p from a single p-variate Skew-Laplace distribution.

    The model:
      1) v_i ~ Gamma( (p+1)/2, scale=2 )
      2) z_i ~ Normal(0, I_p)
      3) x_i = mu + (1 / v_i)*gamma + sqrt(1 / v_i)* L * z_i
         where L = chol(Sigma).

    Parameters
    ----------
    n : int
        Number of samples
    mu : (p,) array
        Location vector
    Sigma : (p,p) array
        Positive-definite covariance
    gamma : (p,) array
        Skewness vector
    random_state : int or None
        If set, seed for reproducibility.

    Returns
    -------
    X : array, shape (n, p)
    """
    # Convert mu, gamma to NumPy arrays (if not already)
    mu = np.asarray(mu, dtype=float)
    gamma = np.asarray(gamma, dtype=float)
    
    rng = np.random.default_rng(random_state)
    p = len(mu)

    # 1) v_i = 1 / Gamma( (p+1)/2, scale=2 )
    # Actually we want v_i = 1 / g_i, where g_i ~ Gamma((p+1)/2, scale=2).
    # => v_i = 1 / g_i
    g_samples = rng.gamma(shape=(p+1)/2.0, scale=2.0, size=n)
    v_samples = 1.0 / g_samples

    # 2) z_i ~ Normal(0, I_p)
    z_samples = rng.normal(size=(n, p))

    # 3) L = chol(Sigma)
    L = cholesky(Sigma)

    # 4) x_i = mu + (1/v_i)*gamma + sqrt(1/v_i)* L * z_i
    X = np.zeros((n, p))
    for i in range(n):
        X[i] = (
            mu
            + v_samples[i] * gamma
            + np.sqrt(v_samples[i]) * (L @ z_samples[i])
        )

    return X


def compute_empirical_info_and_se(model, pi = None, alpha = None, data = None, gamma_res = None, use_model = True):
    """
    Given a fitted SkewLaplaceMix model, compute:
      1) Empirical information matrix I_e = sum_j s_j s_j^T,
      2) Covariance = (I_e)^{-1},
      3) Standard errors = sqrt(diag(Cov)).

    Returns
    -------
    SE : (d,) array
        Standard errors for each parameter in the vector of parameters
        (in the same order we construct the score).
    Cov : (d, d) array
        Approx. covariance matrix = I_e^{-1}.
    """
    # 1) Extract final parameters from the model
    #    pi_k: shape (K,)
    #    alpha_est: list of dicts or list-lists: alpha_est[k] = [mu_k, Sigma_k, gamma_k]
    if use_model:
        pi_est = model.pi_new
    else:
        pi_est = pi
    if use_model:
        alpha_est = model.alpha_new  # length K
    else:
        alpha_est = alpha
        
    

    K = len(pi_est)
    # alpha_est[k][0] => mu_k
    # alpha_est[k][1] => Sigma_k
    # alpha_est[k][2] => gamma_k
    p = alpha_est[0][0].shape[0]

    # 2) Data & final responsibilities
    if use_model:
        X = model.data
    else:
        X = data
    n = X.shape[0]
    # responsibilities => shape (n, K)
    if use_model:
        Z = model.gamma_temp_ar  # e.g. exp(log_gamma_new)
    else:
        Z = gamma_res

    # 3) Compute v1_{i,k}, v2_{i,k}
    V1 = np.zeros((n, K))
    V2 = np.zeros((n, K))

    inv_Sigmas = []
    a_values = []
    for k_i in range(K):
        Sigma_k = alpha_est[k_i][1]
        inv_Sig = np.linalg.inv(Sigma_k)
        inv_Sigmas.append(inv_Sig)

        g_k = alpha_est[k_i][2]
        a_k = np.sqrt(1.0 + g_k @ inv_Sig @ g_k)
        a_values.append(a_k)

    for i in range(n):
        x_i = X[i]
        for k_i in range(K):
            mu_k = alpha_est[k_i][0]
            g_k  = alpha_est[k_i][2]
            invS_k = inv_Sigmas[k_i]
            a_k = a_values[k_i]

            diff = x_i - mu_k
            dist_sq = diff @ invS_k @ diff
            dist = np.sqrt(max(dist_sq, 1e-14))

            # v1_{i,k} = a_k / dist
            # v2_{i,k} = [1 + sqrt(a_k^2 * dist^2)] / a_k^2
            V1[i, k_i] = a_k / dist
            V2[i, k_i] = (1.0 + np.sqrt(a_k*a_k * dist_sq)) / (a_k*a_k)

    # 4) Score function for a single observation
    def score_single_observation(x, z_row, v1_row, v2_row):
        """
        Returns the score vector s_j for one observation j,
        given z_{j,k}, v1_{j,k}, v2_{j,k}, pi_k, alpha[k].
        Order:
          - (K-1) partials wrt pi_0.. pi_{K-2},
          - For each k: partial wrt mu_k, Sigma_k, gamma_k
        """
        score_list = []

        # (A) partial wrt pi_k => eqn(4.2)
        # Each is a scalar => wrap in np.array([ ... ])
        for r in range(K-1):
            val = z_row[r]/pi_est[r] - z_row[K-1]/pi_est[K-1]
            score_list.append(np.array([val]))  # ensure shape (1,)

        # (B) partial wrt mu_k, Sigma_k, gamma_k
        for k_i in range(K):
            mu_k = alpha_est[k_i][0]
            Sigma_k = alpha_est[k_i][1]
            g_k = alpha_est[k_i][2]
            invS_k = inv_Sigmas[k_i]

            diff = x - mu_k

            # eqn(4.3) => partial wrt mu_k => shape (p,)
            s_mu_k = (
                z_row[k_i]*v1_row[k_i]*(invS_k @ diff)
                - z_row[k_i]*v2_row[k_i]*(invS_k @ g_k)
            )

            # eqn(4.4) => partial wrt Sigma_k => shape (p(p+1)/2,)
            GkGkT = np.outer(g_k, g_k)
            outer_diff = np.outer(diff, diff)
            
            A = (
                invS_k
                - v1_row[k_i] * (invS_k @ outer_diff @ invS_k)
                - v2_row[k_i] * (invS_k @ GkGkT   @ invS_k)
            )
            
            
            
            A_diag = np.diag(np.diag(A))          # shape (p,p) with only diagonal of A
            A_corrected = -A + 0.5 * A_diag       # the bracket in eqn. (4.4)
            sSigma_full = z_row[k_i] * A_corrected
            
            # Flatten with vech-lower
            sSigma_k_list = []
            for ii in range(p):
                for jj in range(ii, p):
                    sSigma_k_list.append(sSigma_full[ii, jj])
            sSigma_k = np.array(sSigma_k_list)

            # eqn(4.5) => partial wrt gamma_k => shape (p,)
            s_gamma_k = (
                z_row[k_i]*v1_row[k_i]*(invS_k @ diff)
                - z_row[k_i]*v2_row[k_i]*(invS_k @ g_k)
            )

            score_list.append(s_mu_k)
            score_list.append(sSigma_k)
            score_list.append(s_gamma_k)

        # Now everything in score_list is a 1D array => we can safely concatenate
        return np.concatenate(score_list)

    # 5) Sum up s_j s_j^T over j=1..n
    test_s = score_single_observation(X[0], Z[0], V1[0], V2[0])
    d = len(test_s)

    I_e = np.zeros((d, d))
    for j in range(n):
        s_j = score_single_observation(X[j], Z[j], V1[j], V2[j])
        I_e += np.outer(s_j, s_j)

    # 6) Cov = I_e^{-1}, SE = sqrt(diag(Cov))
    # eigenvalues = np.linalg.eigvalsh(I_e)
    # if np.any(eigenvalues <= 0):
    #     _reg = 1e-6  # Small regularization term
    #     I_e = I_e + _reg * np.eye(I_e.shape[0])
    Cov = np.linalg.pinv(I_e)
    SE = np.sqrt(np.diag(Cov))

    return I_e, SE




def k_means_init_skewlaplace(X, k):
    """
    1) Run KMeans to get cluster assignments.
    2) Compute each cluster's mean, covariance.
    3) gamma_k = 0 (shape vector).
    4) pi_k = #points in cluster / n.
    """
    n, p = X.shape

    # Fit KMeans
    kmeans = KMeans(n_clusters=k, random_state=0).fit(X)
    labels = kmeans.labels_
    centers = kmeans.cluster_centers_  # shape (k, p)

    # pi array
    pi_init = np.zeros(k)
    alpha_init = []
    for cluster_id in range(k):
        # collect points in that cluster
        idx = np.where(labels == cluster_id)[0]
        pi_init[cluster_id] = len(idx) / n

        # mean = cluster center
        mu_k = centers[cluster_id]

        # covariance => empirical from data in cluster
        if len(idx) > 1:
            cov_k = np.cov(X[idx].T, bias=False)  # shape (p,p)
        else:
            # if only one point, fallback to identity or very small
            cov_k = np.eye(p)

        # gamma_k => np.zeros(p)
        gamma_k = np.zeros(p)

        alpha_init.append((mu_k, cov_k, gamma_k))

    return pi_init, alpha_init

def gmm_init_skewlaplace(X, k):
    """
    1) Run GaussianMixture to get cluster assignments.
    2) Extract means, covariances, weights.
    3) gamma_k = 0.
    4) pi_k from GMM.
    """
    n, p = X.shape

    gm = GaussianMixture(n_components=k, random_state=0).fit(X)
    pi_init = gm.weights_  # shape (k,)
    centers = gm.means_    # shape (k, p)
    covariances = gm.covariances_  # shape (k, p, p)

    alpha_init = []
    for cluster_id in range(k):
        mu_k = centers[cluster_id]
        cov_k = covariances[cluster_id]
        gamma_k = np.zeros(p)
        alpha_init.append(( mu_k,cov_k, gamma_k))

    return pi_init, alpha_init

def estimate_alphas_skewlaplace(X, gamma_matrix, alpha_old):
    """
    M-step update for a K-component Multivariate Skew-Laplace mixture,
    using the same formulas as your 2-component MATLAB code but generalized.

    Parameters
    ----------
    X : array, shape (n, p)
        Data matrix.
    gamma_matrix : array, shape (n, K)
        Responsibilities (E-step), gamma_matrix[i, k] = posterior probability that
        sample i belongs to cluster k.
    alpha_old : list of dict, length K
        alpha_old[k] = {"mu": (p,), "Sigma": (p,p), "gamma": (p,)} - old parameters.

    Returns
    -------
    alpha_new : list of dict, length K
        Updated parameter set with structure matching alpha_old.
    """

    n, p = X.shape
    K = len(alpha_old)

    # We will store new parameters in alpha_new
    alpha_new = [None]*K

    # For convenience, precompute inverses and a_k = sqrt(1 + gamma_k^T S_k^-1 gamma_k)
    inv_Sigma_old = []
    a_array = np.zeros(K)
    for k in range(K):
        Sk_inv = np.linalg.inv(alpha_old[k][1])
        inv_Sigma_old.append(Sk_inv)
        gk_old = alpha_old[k][2]
        a_array[k] = np.sqrt(1.0 + gk_old @ Sk_inv @ gk_old)

    # We will accumulate partial sums for each cluster k.
    # For clarity, define arrays of shape (K,) or (K,p) or (K,p,p) as needed.
    sum_zk          = np.zeros(K)       # sum_i gamma_{i,k}
    sum_zk_v1       = np.zeros(K)       # sum_i gamma_{i,k} * v1_{i,k}
    sum_zk_v2       = np.zeros(K)       # sum_i gamma_{i,k} * v2_{i,k}
    sum_zk_x        = np.zeros((K, p))   # sum_i [gamma_{i,k} * x_i]
    sum_zk_v1_x     = np.zeros((K, p))   # sum_i [gamma_{i,k} * v1_{i,k} * x_i]
    # For Sigma:
    # We'll accumulate the numerators S_k_num and the denominators S_k_den:
    S_k_num = [np.zeros((p, p)) for _ in range(K)]
    S_k_den = np.zeros(K)

    # Loop over data points once, computing v1_{i,k}, v2_{i,k} and partial sums.
    for i in range(n):
        xi = X[i]

        for k in range(K):
            gk_old = alpha_old[k][2]   # shape (p,)
            muk_old = alpha_old[k][0]     # shape (p,)
            Sk_inv = inv_Sigma_old[k]
            ak = a_array[k]

            # Distance using old mu, old Sigma:
            diff = xi - muk_old
            dist_sq = diff @ Sk_inv @ diff
            dist_val = np.sqrt(max(dist_sq, 1e-14))  # clip to avoid /0

            # v1_{i,k} and v2_{i,k}:
            v1_ik = ak / dist_val
            # a_k^2 = 1 + gk_old^T S_k^-1 gk_old
            a_sq = ak * ak
            v2_ik = (1.0 + np.sqrt(a_sq * dist_val*dist_val)) / a_sq

            # Accumulate partial sums for gamma_k, mu_k, etc.
            zik = gamma_matrix[i, k]
            sum_zk[k]      += zik
            sum_zk_v1[k]   += zik * v1_ik
            sum_zk_v2[k]   += zik * v2_ik
            sum_zk_x[k]    += zik * xi
            sum_zk_v1_x[k] += zik * v1_ik * xi

            # Now partial sums for Sigma:
            # old formula from MATLAB:
            #   [z_ik * v1_ik * (x_i - mu_k_old)(x_i - mu_k_old)^T]  -  [gk_old*gk_old^T * z_ik*v2_ik]
            # stored in S_k_num
            outer_xdiff = np.outer(diff, diff)
            S_k_num[k] += zik * v1_ik * outer_xdiff  -  np.outer(gk_old, gk_old)* (zik * v2_ik)

            # And the denominator piece:
            #   z_ik*v1_ik * diff^T * Sk_inv * diff  -  (gk_old^T Sk_inv gk_old) * z_ik*v2_ik
            S_k_den[k] += zik * v1_ik * (diff @ Sk_inv @ diff)  \
                          -  (gk_old @ Sk_inv @ gk_old) * (zik * v2_ik)

    # Now finish updates for each cluster k, using the partial sums:
    for k in range(K):
        # (1) Update gamma_k:
        denom_k = sum_zk_v2[k] - (sum_zk[k]**2 / sum_zk_v1[k])
        if abs(denom_k) < 1e-14:
            denom_k = 1e-14 * np.sign(denom_k if denom_k != 0 else 1)
        
        # part_k = sum_zk_x[k] - sum_zk[k]*( sum_zk_v1_x[k]/sum_zk_v1[k] )
        part_k = sum_zk_x[k] - sum_zk[k] * (sum_zk_v1_x[k]/sum_zk_v1[k])
        gk_new = part_k / denom_k  # shape (p,)

        # (2) Update mu_k:
        mu_k_new = (sum_zk_v1_x[k] - gk_new*sum_zk[k]) / sum_zk_v1[k]

        # (3) Update Sigma_k:
        # numerator = p * S_k_num[k]
        # denominator = S_k_den[k], then Sigma_k_new = numerator / denominator
        # again clip if near zero
        denom_sigma_k = S_k_den[k]
        if abs(denom_sigma_k) < 1e-14:
            denom_sigma_k = 1e-14 * np.sign(denom_sigma_k if denom_sigma_k != 0 else 1)
        Sigma_k_new = (p * S_k_num[k]) / denom_sigma_k

        alpha_new[k] = (mu_k_new, Sigma_k_new,gk_new)

    return alpha_new

def compute_log_pdf_skewlaplace(X, mu_k, Sigma_k, gamma_k):
    """
    Computes the log PDF of a multivariate skew-Laplace distribution for given parameters.

    Parameters:
    ----------
    X : array-like, shape (N, p)
        Data points where the density needs to be evaluated.
    mu_k : array-like, shape (p,)
        Mean vector of the distribution.
    Sigma_k : array-like, shape (p, p)
        Covariance matrix (must be positive-definite).
    gamma_k : array-like, shape (p,)
        Skewness vector.

    Returns:
    -------
    log_probs : array, shape (N,)
        Log-density values at each data point.
    """
    N, p = X.shape

    # Invert & logdet Sigma_k
    sign, logdetS = np.linalg.slogdet(Sigma_k)  # sign should be +1 for PD matrix
    invS_k = np.linalg.solve(Sigma_k, np.eye(p))  # More stable than np.linalg.inv()

    # Compute a_k = sqrt(1 + gamma_k.T @ invS_k @ gamma_k)
    gamma_invS_gamma = gamma_k @ invS_k @ gamma_k  # Scalar
    a_k = np.sqrt(1.0 + gamma_invS_gamma)

    # Normalizing constant in log domain
    logC_k = (-0.5 * logdetS
              - (p * np.log(2.0)
                 + (p - 1) * 0.5 * math.log(math.pi)
                 + np.log(a_k)
                 + np.log(gamma_func((p + 1) / 2.0))))

    # Compute Mahalanobis distance
    Xm = X - mu_k  # shape (N, p)
    dist_sq = np.einsum('ij,jk,ik->i', Xm, invS_k, Xm)  # shape (N,)
    dist_val = np.sqrt(np.clip(dist_sq, 1e-14, None))  # Ensure no sqrt(0) errors

    # Compute linear term
    lin_term = np.einsum('ij,j->i', Xm @ invS_k, gamma_k)  # shape (N,)

    # Compute exponent
    exponent = -a_k * dist_val + lin_term

    log_probs = logC_k + exponent
    return log_probs
    

class SkewLaplaceMix(BaseMixture):
    def __init__(self,n_clusters,tol=0.0001,initialization="kmeans",print_log_likelihood=False,max_iter=25, verbose=True):
        super().__init__(n_clusters=n_clusters, EM_type="Soft", mixture_type="identical", tol=tol, print_log_likelihood=print_log_likelihood, max_iter=max_iter, verbose=verbose)
        self.k=n_clusters
        self.initialization=initialization
        self.family = "skew_laplace"
        
    def _log_pdf_skewlaplace(self,X,alphas):
        """
        Returns an array of shape (N, K) where entry (i, k) is
        log [ f_k( X[i] ) ], the log of the k-th Skew-Laplace pdf at X[i].
    
        X: shape (N, p)
        alphas: list of length K, each is {"mu": (p,), "Sigma": (p,p), "gamma": (p,)}.
        """
        N, p = X.shape
        K = len(alphas)
        log_probs = np.zeros((N, K))
    
        for k in range(K):
            mu_k = alphas[k][0]        # shape (p,)
            Sigma_k = alphas[k][1]  # shape (p,p)
            gamma_k = alphas[k][2]  # shape (p,)
    
            log_probs[:, k] = compute_log_pdf_skewlaplace(X, mu_k, Sigma_k, gamma_k)
    
        return log_probs
        
    
    def _estimate_weighted_log_prob_identical(self, X, alpha, pi):
        return self._log_pdf_skewlaplace(X,alpha) + np.log(pi)
    
    def fit(self,sample):
        start_time = time.time()
        self.data = self._process_data(sample)
        self.n, self.p = self.data.shape
        self.total_parameters = (self.k -1) + self.k* (2*self.p + (self.p*(self.p+1)/2))
        np.random.seed(0)
        if self.initialization == "kmeans":
            self.pi_not, self.alpha_not = k_means_init_skewlaplace(self.data, self.k)
        elif self.initialization == "gmm":
            self.pi_not, self.alpha_not = gmm_init_skewlaplace(self.data, self.k)
        self.alpha_temp = self.alpha_not
        self.pi_temp = self.pi_not
        pi_new,alpha_new, log_likelihood_new,log_gamma_new=self._fit(self.data,self.pi_temp,self.alpha_temp,estimate_alphas_skewlaplace)
        self.pi_new = pi_new
        self.alpha_new = alpha_new
        self.cluster = log_gamma_new.argmax(axis=1)
        self.gamma_temp_ar = np.exp(log_gamma_new)
        self.log_likelihood_new = log_likelihood_new
        if self.verbose:
            print("Mixtures of Skew-Laplace Fitting Done Successfully")
        end_time = time.time()
        self.execution_time=end_time-start_time
    
    
    
    
    
    def get_params(self):
        #print("The estimated pi values are ", self.pi_new)
        #print("The estimated alpha values are ", self.alpha_new)

        return self.pi_new, self.alpha_new

    def predict(self):
        return self.cluster

    def predict_new(self, x):
        x=self._process_data(x)
        data_lol = x.tolist()
        cluster, _ = mixture_clusters(self.gamma_temp_ar, data_lol)

        return cluster


    def responsibilities(self):

        return self.gamma_temp_ar

    def n_iter(self):
        return len(self.log_likelihoods)
    
    def get_info_mat(self):
        
        IM, SE = compute_empirical_info_and_se(self)
        
        return IM, SE
        



import numpy as np
from numpy.linalg import inv, det
from scipy.linalg import sqrtm
from scipy.stats import norm
from scipy.integrate import quad
from scipy.optimize import minimize_scalar
from scipy.special import gamma as gamma_func
import time
from fmvmm.mixtures._base import BaseMixture
from sklearn.cluster import KMeans
from fmvmm.utils.utils_mixture import (mixture_clusters)
from fmvmm.mixsmsn.information_matrix_smsn import info_matrix, standard_errors_from_info_matrix


def dmvSS(X, mu, Sigma, lam, nu):
    """
    Multivariate Slash (or Skew-Slash) density for each row of X, 
    as in your prior code. If lam=0, it becomes the standard Slash distribution.

    X  : (n, p)
    mu : (p,)
    Sigma : (p, p)
    lam : (p,)  (for standard slash, lam=0)
    nu : scalar slash parameter
    Returns: (n,) array of PDF values
    """

    X = np.asarray(X)
    mu = np.asarray(mu)
    lam = np.asarray(lam)
    n, p = X.shape

    # Safeguard for near-singular Sigma
    det_Sig = np.linalg.det(Sigma)
    if det_Sig < 1e-300:
        Sigma = Sigma + 1e-6*np.eye(p)
        det_Sig = np.linalg.det(Sigma)

    inv_Sig = inv(Sigma)
    diff = X - mu
    dj = np.sum((diff @ inv_Sig)*diff, axis=1)  # Mahalanobis distances

    # factor for lam != 0 (Skew-Slash), lam=0 => factor is zero => pnorm(0)=0.5
    sqrt_Sig = sqrtm(Sigma + 1e-12*np.eye(p))
    inv_sqrt_Sig = inv(sqrt_Sig)
    factor = lam @ inv_sqrt_Sig  # shape (p,)
    A_vals = np.einsum('j,ij->i', factor, diff)  # shape (n,)

    pdf_vals = np.zeros(n)

    
    for i in range(n):
        dj_i = dj[i]
        A_i = A_vals[i]

        # Evaluate dmvSS by direct integral from 0..1
        def integrand(u):
            return (2.0*nu*(u**(nu-1.0))
                    * ((u/(2.0*np.pi))**(p/2.0))/np.sqrt(det_Sig)
                    * np.exp(-0.5*u*dj_i)
                    * norm.cdf(np.sqrt(u)*A_i))

        val, _ = quad(integrand, 0.0, 1.0, limit=50)
        pdf_vals[i] = val

    return pdf_vals

def d_mixedmvSS(X, pi_list, mu_list, Sigma_list, lam_list, nu):
    """
    Mixture of multivariate Slash (or Skew-Slash) distributions, 
    all sharing the same slash parameter nu, possibly lam=0 for each cluster 
    for pure slash.

    pi_list : length g
    mu_list, Sigma_list, lam_list : each length g
    nu : scalar
    Returns: (n,) array of mixture pdf values
    """
    g = len(pi_list)
    n = X.shape[0]
    total = np.zeros(n)
    for j in range(g):
        dens_j = dmvSS(X, mu_list[j], Sigma_list[j], lam_list[j], nu)
        total += pi_list[j]*dens_j
    return total

def estimate_alphas_slash(
    X,
    gamma,                   # responsibilities (n, g)
    alpha_prev,              # list of (mu_j, Sigma_j, shape_j=0, nu)
    pi_prev,                 # (g,)
    dmvSS_func,
    d_mixedmvSS_func,
    bounds_nu=(1e-4, 100.0),
    **kwargs
):
    """
    Single M-step update for a mixture of (non-skew) Slash components.
    shape=0 => symmetrical slash distribution.

    Returns
    -------
    alpha_new : list of length g => (mu_j, Sigma_j, shape_j=0, nu_new)
    """
    X = np.asarray(X)
    n, p = X.shape
    g = len(alpha_prev)

    # old nu (single shared)
    nu_old = alpha_prev[0][3]

    # Extract old parameters
    mu_old_list = [alpha_prev[j][0] for j in range(g)]
    Sigma_old_list = [alpha_prev[j][1] for j in range(g)]
    # shape_j=0 for slash => do not need a real shape

    # We'll compute E-step-like quantities: u[i] and E[i].
    # Because shape=0 => Delta=0 => A=0 => pnorm(0)=0.5 in the integral, etc.

    # We'll accumulate S1, S2, S3
    S1 = np.zeros((n, g))
    S2 = np.zeros((n, g))
    S3 = np.zeros((n, g))

    from numpy.linalg import inv
    from scipy.linalg import sqrtm
    from scipy.stats import norm
    from scipy.integrate import quad

    for j in range(g):
        mu_j = mu_old_list[j]
        Sigma_j = Sigma_old_list[j]

        # shape=0 => Delta=0 => Gama_j = Sigma_j
        # Mtij2 => 1/(1 + 0) => 1 => Mtij=1 => mutij[i]=0 => A=0
        # So a lot of the E-step formulas simplify.

        # pdf_j = dmvSS(X, mu_j, Sigma_j, shape=0, nu_old)
        pdf_j = dmvSS_func(X, mu_j, Sigma_j, np.zeros(p), nu_old)
        pdf_j = np.maximum(pdf_j, 1e-300)

        
        # 1) Compute dj
        inv_Sigma_j = inv(Sigma_j + 1e-12*np.eye(p))
        diff = X - mu_j
        dj_arr = np.sum((diff @ inv_Sigma_j)*diff, axis=1)

        # 2) Compute u[i]
        
        det_Sig_j = max(1e-300, np.linalg.det(Sigma_j))
        c_val_u = (nu_old*(2.0**(1 - p/2.0))*(np.pi**(-0.5*p))*(det_Sig_j**(-0.5)))

        def integrand_u(u, di):
            return u**(nu_old + p/2.0)*np.exp(-0.5*u*di)*0.5  # 0.5 from pnorm(0)

        u_arr = np.zeros(n)
        for i in range(n):
            di = dj_arr[i]
            val_int, _ = quad(lambda u: integrand_u(u, di), 0.0, 1.0, limit=50)
            u_arr[i] = (c_val_u*val_int) / pdf_j[i]

        # 3) Compute E[i]
        
        from scipy.stats import gamma as gamma_dist
        cE_num = (2.0**(nu_old+1))*nu_old*gamma_func((2*nu_old+p+1)/2.0)
        cE_den = ((np.pi)**(0.5*(p+1)))*np.sqrt(det_Sig_j)
        cE = cE_num/(cE_den + 1e-300)

        E_arr = np.zeros(n)
        for i in range(n):
            val_pdf = pdf_j[i]
            di = dj_arr[i]
            
            shape_ = (2*nu_old+p+1)/2.0
            rate_  = di/2.0
            cdf_val = gamma_dist.cdf(1.0, a=shape_, scale=1.0/(rate_+1e-15))

            power_factor = (di**(-(2*nu_old+p+1)/2.0)) if di>0 else 1e+15  # caution if di=0
            E_i = (cE*power_factor*cdf_val)/(val_pdf+1e-300)
            E_arr[i] = E_i

        # 4) S1, S2, S3 with shape=0 => Delta=0 => mutij=0 => Mtij=1 => 
        
        gam_j = gamma[:, j]
        S1[:, j] = gam_j*u_arr
        S2[:, j] = gam_j*E_arr
        S3[:, j] = gam_j  # since shape=0 => that formula is 1

    # Now update parameters:
    alpha_new_partial = []
    for j in range(g):
        mu_j_old = mu_old_list[j]
        sum_S1_j = np.sum(S1[:, j])
        sum_S2_j = np.sum(S2[:, j])
        sum_S3_j = np.sum(S3[:, j])

        
        if sum_S1_j < 1e-15:
            mu_j_new = mu_j_old
        else:
            sum_S1_X = np.einsum('i,ij->j', S1[:, j], X)
            mu_j_new = sum_S1_X/(sum_S1_j + 1e-300)

        
        dif_new = X - mu_j_new
        sum2 = np.zeros((p,p))
        for i in range(n):
            sum2 += S1[i, j]*np.outer(dif_new[i], dif_new[i])
        n_j = np.sum(gamma[:, j])
        Gama_j_new = sum2/(n_j + 1e-300)
        Sigma_j_new = Gama_j_new  # shape=0 => Delta=0 => Sigma=Gama

        # shape_j_new=0 => forcibly zero
        shape_j_new = np.zeros(p)

        alpha_new_partial.append((mu_j_new, Sigma_j_new, shape_j_new, nu_old))

    # Finally, update nu by maximizing the mixture log-likelihood
    def neg_loglik_slash(nu_val):
        mu_list = [a[0] for a in alpha_new_partial]
        Sig_list= [a[1] for a in alpha_new_partial]
        lam_list= [a[2] for a in alpha_new_partial]  # all zeros
        pdf_mix = d_mixedmvSS_func(X, pi_prev, mu_list, Sig_list, lam_list, nu_val)
        pdf_mix = np.maximum(pdf_mix, 1e-300)
        return -np.sum(np.log(pdf_mix))

    res = minimize_scalar(
        neg_loglik_slash,
        bounds=bounds_nu,
        method='bounded',
        options={"xatol":1e-6}
    )
    nu_new = res.x

    alpha_new = []
    for j in range(g):
        mu_j, Sig_j, shp_j, _ = alpha_new_partial[j]
        alpha_new.append((mu_j, Sig_j, shp_j, nu_new))

    return alpha_new

def k_means_init_slash(data, k):
    """
    K-means initialization for a mixture of (non-skew) Slash distributions.
    shape=0 by default.
    """
    data = np.asarray(data)
    n, p = data.shape

    kmeans = KMeans(n_clusters=k, random_state=0)
    kmeans.fit(data)

    mu_init = kmeans.cluster_centers_  # (k,p)

    # group data
    data_cwise = [[] for _ in range(k)]
    for i in range(n):
        data_cwise[kmeans.labels_[i]].append(data[i])
    data_cwise = [np.asarray(arr).T for arr in data_cwise]

    pi_init = []
    alpha_init = []
    # We pick a default slash parameter, e.g. nu=2
    for j in range(k):
        cluster_data = data_cwise[j]
        size_j = cluster_data.shape[1]
        pi_init.append(size_j/n)

        if size_j <= 1:
            Sigma_j = np.eye(p)
        else:
            Sigma_j = np.cov(cluster_data)

        # shape=0
        shape_j = np.zeros(p)
        nu_j = 2.0

        alpha_init.append((mu_init[j], Sigma_j, shape_j, nu_j))

    return np.array(pi_init), alpha_init


class SlashMix(BaseMixture):
    """
    Mixture of Multivariate Slash (non-skew) distributions,
    sharing a single slash parameter 'nu'.
    shape=0 => symmetrical slash.
    """

    def __init__(self,
                 n_clusters,
                 tol=1e-4,
                 initialization="kmeans",
                 print_log_likelihood=False,
                 max_iter=25,
                 verbose=True):
        super().__init__(n_clusters=n_clusters,
                         EM_type="Soft",
                         mixture_type="identical",
                         tol=tol,
                         print_log_likelihood=print_log_likelihood,
                         max_iter=max_iter,
                         verbose=verbose)
        self.k = n_clusters
        self.initialization = initialization
        self.family = "slash"

    def _log_pdf_slash(self, X, alphas):
        """
        Evaluate log-pdf of each cluster's Slash distribution at each row of X.
        alphas[j] = (mu_j, Sigma_j, shape_j=0, nu_j)
        Returns: (n, g) array of log densities
        """
        N, p = X.shape
        g = len(alphas)
        logvals = np.empty((N, g))
        for j in range(g):
            mu_j, Sigma_j, shape_j, nu_j = alphas[j]
            # shape_j=0 => standard slash
            pdf_j = dmvSS(X, mu_j, Sigma_j, shape_j, nu_j)
            pdf_j = np.maximum(pdf_j, 1e-300)
            logvals[:, j] = np.log(pdf_j)
        return logvals

    def _estimate_weighted_log_prob_identical(self, X, alpha, pi):
        """
        Weighted log-prob for identical mixture type: log f_j(X) + log pi_j
        """
        log_pdf = self._log_pdf_slash(X, alpha)  # (n, g)
        log_pi = np.log(pi)                      # (g,)
        return log_pdf + log_pi

    def fit(self, sample):
        start_time = time.time()
        self.data = self._process_data(sample)
        self.n, self.p = self.data.shape
        self.total_parameters = (self.k -1) + self.k* (self.p + (self.p*(self.p+1)/2) +1)

        # Initialization
        if self.initialization == "kmeans":
            self.pi_not, self.alpha_not = k_means_init_slash(self.data, self.k)
        else:
            raise NotImplementedError("Only 'kmeans' initialization is supported for SlashMix.")

        self.alpha_temp = self.alpha_not
        self.pi_temp    = self.pi_not

        # Provide references for M-step
        def dmvSS_local(data, mu, Sigma, lam, nu):
            return dmvSS(data, mu, Sigma, lam, nu)
        def d_mixedmvSS_local(data, pi_list, mu_list, Sigma_list, lam_list, nu):
            return d_mixedmvSS(data, pi_list, mu_list, Sigma_list, lam_list, nu)

        self.dmvSS_func = dmvSS_local
        self.d_mixedmvSS_func = d_mixedmvSS_local

        # Run the base-class EM
        pi_new, alpha_new, log_likelihood_new, log_gamma_new = self._fit(
            self.data,
            self.pi_temp,
            self.alpha_temp,
            estimate_alphas_function=self._estimate_alphas_wrapper
        )

        self.pi_new = pi_new
        self.alpha_new = alpha_new
        self.cluster = log_gamma_new.argmax(axis=1)
        self.gamma_temp_ar = np.exp(log_gamma_new)
        self.log_likelihood_new = log_likelihood_new

        if self.verbose:
            print("Mixtures of Slash Fitting Done Successfully!")
        end_time = time.time()
        self.execution_time = end_time - start_time

    def _estimate_alphas_wrapper(self, X, gamma, alpha_prev, **kwargs):
        """
        A wrapper for the M-step: calls estimate_alphas_slash once.
        """
        pi_prev = self.pi_temp
        alpha_new = estimate_alphas_slash(
            X,
            gamma,
            alpha_prev,
            pi_prev,
            dmvSS_func=self.dmvSS_func,
            d_mixedmvSS_func=self.d_mixedmvSS_func,
            **kwargs
        )
        return alpha_new
    
    def get_params(self):
        #print("The estimated pi values are ", self.pi_new)
        #print("The estimated alpha values are ", self.alpha_new)

        return self.pi_new, self.alpha_new

    def predict(self):
        return self.cluster

    def predict_new(self, x):
        x=self._process_data(x)
        data_lol = x.tolist()
        cluster, _ = mixture_clusters(self.gamma_temp_ar, data_lol)

        return cluster


    def responsibilities(self):

        return self.gamma_temp_ar

    def n_iter(self):
        return len(self.log_likelihoods)

    def get_info_mat(self):
        
        im = info_matrix(self)
        se = standard_errors_from_info_matrix(im)
        
        return im, se



"""
Most of the codes are used from git repository mvem. Huge thanks to the
devoloper. 

"""

import numpy as np
from fmvmm.distributions import multivariate_genhyperbolic as ghypmv

def logpdf(x, chi, psi, mu, sigma, gamma):
    """
    Log-probability density function of the multivariate hyperbolic distribution.
    We use the (lmbda, chi, psi, mu, sigma, gamma)-parameterisation.

    :param x: An array of shape (n, p) containing n observations of some
        p-variate data with n > p.
    :type x: np.ndarray
    :param chi: Univariate parameter > 0.
    :type chi: float
    :param psi: Univariate parameter > 0.
    :type psi: float
    :param mu: Location parameter with shape (p,).
    :type mu: np.ndarray
    :param sigma: A positive semi-definite array with shape (p, p).
    :type sigma: np.ndarray
    :param gamma: Parameter with shape (p,).
    :type gamma: np.ndarray
    :return: The log-density at each observation.
    :rtype: np.ndarray with shape (n,).
    """
    lmbda = (len(mu)+1)/2
    return ghypmv.logpdf(x, lmbda, chi, psi, mu, sigma, gamma)

def pdf(x, chi, psi, mu, sigma, gamma):
    """
    Probability density function of the multivariate hyperbolic distribution.
    We use the (lmbda, chi, psi, mu, sigma, gamma)-parameterisation.

    :param x: An array of shape (n, p) containing n observations of some
        p-variate data with n > p.
    :type x: np.ndarray
    :param chi: Univariate parameter > 0.
    :type chi: float
    :param psi: Univariate parameter > 0.
    :type psi: float
    :param mu: Location parameter with shape (p,).
    :type mu: np.ndarray
    :param sigma: A positive semi-definite array with shape (p, p).
    :type sigma: np.ndarray
    :param gamma: Parameter with shape (p,).
    :type gamma: np.ndarray
    :return: The density at each observation.
    :rtype: np.ndarray with shape (n,).
    """
    lmbda = (len(mu)+1)/2
    return ghypmv.pdf(x, lmbda, chi, psi, mu, sigma, gamma)

def loglike(x, chi, psi, mu, sigma, gamma):
    """
    Log-likelihood function of the multivariate hyperbolic distribution.
    We use the (lmbda, chi, psi, mu, sigma, gamma)-parameterisation.

    :param x: An array of shape (n, p) containing n observations of some
        p-variate data with n > p.
    :type x: np.ndarray
    :param chi: Univariate parameter > 0.
    :type chi: float
    :param psi: Univariate parameter > 0.
    :type psi: float
    :param mu: Location parameter with shape (p,).
    :type mu: np.ndarray
    :param sigma: A positive semi-definite array with shape (p, p).
    :type sigma: np.ndarray
    :param gamma: Parameter with shape (p,).
    :type gamma: np.ndarray
    :return: The log-likelihood for given all observations and parameters.
    :rtype: float
    """
    return np.sum(logpdf(x, chi, psi, mu, sigma, gamma))

def total_params(chi, psi, mu, sigma, gamma):
    p = len(mu)
    
    return 2 + 2*p + (p*(p+1)/2)

def rvs(chi, psi, mu, sigma, gamma, size):
    """
    Random number generator of the multivariate hyperbolic distribution.
    We use the (lmbda, chi, psi, mu, sigma, gamma)-parameterisation.

    :param chi: Univariate parameter > 0.
    :type chi: float
    :param psi: Univariate parameter > 0.
    :type psi: float
    :param mu: Location parameter with shape (p,).
    :type mu: np.ndarray
    :param sigma: A positive semi-definite array with shape (p, p).
    :type sigma: np.ndarray
    :param gamma: Parameter with shape (p,).
    :type gamma: np.ndarray
    :param size: The number of samples to draw.
    :type size: int
    :return: The random p-variate numbers generated.
    :rtype: np.ndarray with shape (n, p).
    """
    lmbda = (len(mu)+1)/2
    return ghypmv.rvs(lmbda, chi, psi, mu, sigma, gamma, size)

def mean(chi, psi, mu, sigma, gamma):
    """
    Mean function of the multivariate hyperbolic distribution. We use
    the (lmbda, chi, psi, mu, sigma, gamma)-parameterisation.

    :param chi: Univariate parameter > 0.
    :type chi: float
    :param psi: Univariate parameter > 0.
    :type psi: float
    :param mu: Location parameter with shape (p,).
    :type mu: np.ndarray
    :param sigma: A positive semi-definite array with shape (p, p).
    :type sigma: np.ndarray
    :param gamma: Parameter with shape (p,).
    :type gamma: np.ndarray
    :return: The mean of the specified distribution.
    :rtype: np.ndarray with shape (p,).
    """
    lmbda = (len(mu)+1)/2
    return ghypmv.mean(lmbda, chi, psi, mu, sigma, gamma)

def var(chi, psi, mu, sigma, gamma):
    """
    Variance function of the multivariate hyperbolic distribution. We use
    the (lmbda, chi, psi, mu, sigma, gamma)-parameterisation.

    :param chi: Univariate parameter > 0.
    :type chi: float
    :param psi: Univariate parameter > 0.
    :type psi: float
    :param mu: Location parameter with shape (p,).
    :type mu: np.ndarray
    :param sigma: A positive semi-definite array with shape (p, p).
    :type sigma: np.ndarray
    :param gamma: Parameter with shape (p,).
    :type gamma: np.ndarray
    :return: The variance of the specified distribution.
    :rtype: np.ndarray with shape (n,).
    """
    lmbda = (len(mu)+1)/2
    return ghypmv.var(lmbda, chi, psi, mu, sigma, gamma)

def fit(x, alpha_bar=1, symmetric=False, standardize=False, nit=2000, reltol=1e-8,
        abstol=1e-7, silent=False, fmu=None, fsigma=None, fgamma=None, return_loglike=False):
    """
    Estimate the parameters of the hyperbolic distribution. We use the
    (lmbda, chi, psi, mu, sigma, gamma)-parameterisation.

    :param x: An array of shape (n, p) containing n observations of some
        p-variate data with n > p.
    :type x: np.ndarray
    :param alpha_bar: The initial value of alpha_bar, a positive real number,
        where alpha_bar = chi =  psi.
    :type alpha_bar: float, optional
    :param symmetric: Whether to fit a symmetric distribution or not. Default
        to False.
    :type symmetric: bool, optional
    :param standardize: Whether to standardize the data before fitting or not.
        Default to False.
    :type standardize: bool, optional
    :param nit: The maximum number of iterations to use in the EM algorithm.
        Defaults to 2000.
    :type nit: int, optional
    :param reltol: The relative convergence criterion for the log-likelihood
        function. Defaults to 1e-8.
    :type reltol: float, optional
    :param abstol: The relative convergence criterion for the log-likelihood
        function. Defaults to 1e-7.
    :type abstol: float, optional
    :param silent: Whether to print the log-likelihoods and parameter estimates
        during fitting or not. Defaults to False.
    :type silent: bool, optional
    :param fmu: If fmu!=None, force mu to fmu. Defaults to None.
    :type fmu: np.ndarray, optional
    :param fsigma: If fsigma!=None, force sigma to fsigma. Defaults to None.
    :type fsigma: np.ndarray, optional
    :param fgamma: If fgamma!=None, force gamma to fgamma. Defaults to None.
    :type fgamma: np.ndarray, optional
    :param return_loglike: Return a list of log-likelihood values at each iteration.
        Defaults to False.
    :type return_loglike: np.ndarray, optional
    :return: The fitted parameters (<float> chi, <float> psi, <array> mu, <array> sigma,
        <array> gamma). Also returns a list of log-likelihood values at each iteration
        of the EM algorithm if ``return_loglike=True``.
    :rtype: tuple
    """

    opt_pars = {"lmbda": False, "alpha_bar": True, "mu": fmu is None,
                "sigma": fsigma is None, "gamma": fgamma is None}
    lmbda = (x.shape[1]+1)/2
    fit = ghypmv.fitghypmv(
        x, lmbda=lmbda, alpha_bar=alpha_bar, mu=fmu, sigma=fsigma, gamma=fgamma,
        symmetric=symmetric, standardize=standardize, nit=nit, reltol=reltol,
        abstol=abstol, silent=silent, opt_pars=opt_pars)

    chi, psi = ghypmv._alphabar2chipsi(fit["alpha_bar"], fit["lmbda"])
    if return_loglike:
        return fit["lmbda"], chi, psi, fit["mu"], fit["sigma"], fit["gamma"], fit["ll"]
    return chi, psi, fit["mu"], fit["sigma"], fit["gamma"]

def info_mat(x, chi, psi, mu, sigma, gamma):
    from fmvmm.utils.utils_fmm import compute_info_scipy_fmvmm
    
    IM = compute_info_scipy_fmvmm(logpdf,[np.array([chi]), np.array([psi]), mu, sigma, gamma],x)
    
    return IM


from scipy.optimize import approx_fprime
import numpy as np
from fmvmm.utils.utils_fmm import (fmm_kmeans_init,fmm_gmm_init, fmm_loglikelihood, fmm_responsibilities,
                             fmm_pi_estimate, fmm_estimate_alphas, fmm_aic, fmm_bic, fmm_icl)


from fmvmm.utils.utils_mixture import (mixture_clusters)
import math

import itertools

from fmvmm.distributions import multivariate_skewnorm as mvsn
from fmvmm.distributions import multivariate_genhyperbolic as mghp
from fmvmm.distributions import multivariate_genskewt as mgst
from fmvmm.distributions import multivariate_hyperbolic as mvhb
from fmvmm.distributions import multivariate_norm as mvn
from fmvmm.distributions import multivariate_norminvgauss as mnig
from fmvmm.distributions import multivariate_t as mvt
from fmvmm.distributions import multivariate_vargamma as mvvg
from fmvmm.distributions import multivariate_skew_laplace as mvsl
from fmvmm.distributions import multivariate_skew_t_smsn as mvst
from fmvmm.distributions import multivariate_skewnorm_cont as msnc
from fmvmm.distributions import multivariate_skewslash as mssl
from fmvmm.distributions import multivariate_slash as msl
import traceback
from scipy.special import logsumexp
from fmvmm.mixtures._base import BaseMixture
from fmvmm.utils.utils_dmm import hard_assignments, mixture_counts, mixture_proportions_info
from fmvmm.mixtures.mixmgh import approx_hessian_scipy
import warnings
warnings.filterwarnings('ignore')

dist_map = {"mvsn": mvsn, "mghp": mghp, "mgst": mgst, "mvhb": mvhb,
            "mvn": mvn, "mnig": mnig, "mvt": mvt, "mvvg": mvvg, "mvsl": mvsl,
            "mvst": mvst, "msnc": msnc, "mssl": mssl, "msl": msl}

all_dist = ["mvsn", "mghp", "mgst", "mvhb", "mvn", "mnig", "mvt", "mvvg",
            "mvsl", "mvst", "msnc","mssl","msl"]

def convert_to_numpy(tuples_list):
    """
    Given a list of tuples, convert any integer or float inside the tuples 
    into a NumPy array of shape (1,).
    """
    def convert_item(item):
        if isinstance(item, (int, float)):
            return np.array([item])
        elif isinstance(item, tuple):
            return tuple(convert_item(sub_item) for sub_item in item)
        elif isinstance(item, list):  # If lists are inside tuples, handle them too
            return [convert_item(sub_item) for sub_item in item]
        else:
            return item  # Keep other types unchanged

    return [tuple(convert_item(item) for item in tpl) for tpl in tuples_list]


def compute_info_individual_fmvmm(dist_comp, params_init, X, epsilon=1e-6):
    """
    Use SciPy to approximate the Hessian of the negative log-likelihood 
    for the GH mixture, then invert to get the empirical info matrix.

    Returns: I_e, Cov, SE
    """
    IM = dist_comp.info_mat(X, *params_init)
    
    # Flatten final parameters
    n, p = X.shape
    param_shapes = [np.array(p).shape for p in params_init]
    theta_hat = flatten_params(params_init)

    d = len(theta_hat)

    

    return IM, len(theta_hat)

def organize_data_by_clusters(data, cluster_predictions):
    # Get the unique cluster labels
    unique_clusters = np.unique(cluster_predictions)
    
    # Initialize a list to store data points for each cluster
    data_by_clusters = [np.empty((0, data.shape[1])) for _ in unique_clusters]
    
    # Iterate over each data point and its cluster prediction
    for point, cluster in zip(data, cluster_predictions):
        # Append the data point to the corresponding cluster
        data_by_clusters[cluster] = np.vstack((data_by_clusters[cluster], point))
    
    return data_by_clusters


def flatten_params(params):
    """
    Flatten parameters into a single vector.
    
    Parameters:
        params (tuple): The parameters.
        
    Returns:
        flat_params (ndarray): The flattened parameters.
    """
    return np.array(np.concatenate([p.flatten() if isinstance(p, np.ndarray) else np.array([p]) for p in params]), dtype=float)


def reshape_params(flat_params, param_shapes):
    """
    Reshape a flattened parameter vector into original shapes.
    
    Parameters:
        flat_params (ndarray): The flattened parameters.
        param_shapes (list): The shapes of the original parameters.
        
    Returns:
        params (tuple): The reshaped parameters.
    """
    params = []
    start = 0
    for shape in param_shapes:
        end = start + np.prod(shape)
        params.append(np.reshape(flat_params[start:end], shape))
        start = end
    return tuple(params)

def ensure_positive_diagonal(sum_outer):
    """
    Ensure that the matrix has positive diagonal elements.
    
    Parameters:
        sum_outer (ndarray): The matrix. Shape (num_params, num_params).
        
    Returns:
        sum_outer_pos (ndarray): The matrix with positive diagonal elements.
    """
    # Perform eigenvalue decomposition
    eigenvalues, eigenvectors = np.linalg.eig(sum_outer)
    
    # Adjust eigenvalues to ensure positivity
    eigenvalues[eigenvalues < 0] = 1e-6  # Set negative eigenvalues to a small positive value
    
    # Reconstruct the matrix
    sum_outer_pos = np.dot(eigenvectors, np.dot(np.diag(eigenvalues), np.linalg.inv(eigenvectors)))
    
    return sum_outer_pos
    
def ecdf_from_log_density(log_density):
    """
    Compute the empirical cumulative distribution function (ECDF) from log-density values.

    Parameters:
        log_density (numpy.ndarray): The array of log-density values.

    Returns:
        numpy.ndarray: The empirical cumulative distribution function values.
    """
    # Convert log-density to density by exponentiation
    density = np.exp(log_density)

    # Sort the density values in ascending order
    sorted_density = np.sort(density)

    # Compute the cumulative sum of the sorted density values
    cumulative_sum = np.cumsum(sorted_density)

    # Normalize the cumulative sum to get ECDF values
    ecdf_values = cumulative_sum / np.sum(density)

    return ecdf_values


def fmm_sorted_lpdf_cdf(pi_temp,alpha_temp,data_lol,list_of_dist):
    n = len(data_lol)
    k = len(alpha_temp)
    dist_variables = [dist_map[list_of_dist[j]]
                           for j in range(len(list_of_dist))]
    dist_comb = list(
        itertools.combinations(dist_variables, k))[0]
    log_likelihood_values_temp = []
    for c in range(n):
        try:
            log_likelihood_old_temp = math.log(np.nansum(
                [pi_temp[f]*dist_comb[f].pdf(np.reshape(np.array(data_lol[c]),(1,len(data_lol[c]))), *alpha_temp[f]) for f in range(k)]))
        except:
            log_likelihood_old_temp = math.log1p(np.nansum(
                [pi_temp[f]*dist_comb[f].pdf(np.reshape(np.array(data_lol[c]),(1,len(data_lol[c]))), *alpha_temp[f]) for f in range(k)]))
        log_likelihood_values_temp.append(log_likelihood_old_temp)
    
    sorted_lists = sorted(zip(log_likelihood_values_temp, data_lol),reverse=False)
    sorted_lpdf, sorted_data = zip(*sorted_lists)
    
    
    cdfs=ecdf_from_log_density(log_likelihood_values_temp)
    return sorted_lpdf,cdfs

def add_elements_at_indices(my_list, indices_to_add, new_elements):
    for i, index in enumerate(indices_to_add):
        my_list.insert(index+i, new_elements[i])
    return my_list

def flatten(seq):
    flat_list = []
    for item in seq:
        if isinstance(item, (list, tuple, np.ndarray)):  # Check if item is a list, tuple or ndarray
            flat_list.extend(flatten(item))  # Recursively call flatten on the nested list, tuple or ndarray
        else:
            flat_list.append(item)  # Append non-list/tuple/ndarray items to the flat list
    return flat_list


def remove_nan_tuples(list_of_lists, list_of_list_of_tuples):
    # Initialize a list to store indices of sub-lists or tuples to remove
    indices_to_remove = []

    for i in range(len(list_of_list_of_tuples)):
        temp=flatten(list_of_list_of_tuples[i])
        if any(np.isnan(temp)) or any(np.isinf(temp)):
            indices_to_remove.append(i)
            
    for i, inner_list in enumerate(list_of_lists):
        if any(val == 0 for val in inner_list):
            indices_to_remove.append(i)
            

    # Remove sub-lists and tuples using the stored indices
    if len(indices_to_remove)>0:
        list_of_lists = [list_of_lists[i] for i in range(len(list_of_lists)) if i not in indices_to_remove]
        list_of_list_of_tuples = [list_of_list_of_tuples[i] for i in range(len(list_of_list_of_tuples)) if i not in indices_to_remove]
        
    return list_of_lists, list_of_list_of_tuples, indices_to_remove

def remove_elements_by_index(my_list, index_list):
    if len(index_list)>0:
        return [item for i, item in enumerate(my_list) if i not in index_list]
    else:
        return my_list

def get_dist_names(dist_comb):
    return [str(c.__name__) for c in dist_comb]




class fmvmm(BaseMixture):
    def __init__(self,n_clusters,tol=0.0001, list_of_dist=all_dist,specific_comb=False,initialization="kmeans",print_log_likelihood=False,max_iter=25, verbose=True, debug = False):
        super().__init__(n_clusters=n_clusters, EM_type="Hard", mixture_type="nonidentical", tol=tol, print_log_likelihood=print_log_likelihood, max_iter=max_iter, verbose=verbose)
        self.k=n_clusters
        self.list_of_dist = list_of_dist
        self.specific_comb=specific_comb
        self.dist_variables = [dist_map[list_of_dist[j]]
                               for j in range(len(list_of_dist))]
        self.specific_comb=specific_comb
        if self.specific_comb==True:
            self.dist_combs = list(
                itertools.combinations(self.dist_variables, self.k))
        else:
            self.dist_combs = list(
                itertools.combinations_with_replacement(self.dist_variables, self.k))
        self.initialization=initialization
        self.debug = debug
    
    def _log_pdf_non_identical(self,X,alphas,dist_comb):
        N,p=X.shape
        k=len(alphas)
        probs=np.empty((N, k))
        for j in range(k):
            alpha=alphas[j]
            probs[:, j]=dist_comb[j].logpdf(X,*alpha)
        return probs    
    
    def _estimate_weighted_log_prob_nonidentical(self, X, alpha, pi, dist_comb):
        return self._log_pdf_non_identical(X,alpha,dist_comb) + np.log(pi)
    
    def fit(self,sample):
        self.data = self._process_data(sample)
        self.n=len(sample)
        # self.sample=self._process_data(sample)
        self.list_aic = []
        self.list_bic = []
        self.list_icl = []
        self.list_pi = []
        self.list_alpha = []
        self.list_cluster = []
        self.list_log_likelihood = []
        self.list_all_log_likelihood = []
        self.list_gamma_matrix = []
        self.not_worked_dist=[]
        self.worked_dist=[]
        for l in range(len(self.dist_combs)):
            try:
                
                if self.initialization=="kmeans":
                    
                    self.pi_not, self.alpha_not = fmm_kmeans_init(
                        self.data, self.k, self.dist_combs[l])
                    self.alpha_temp = self.alpha_not
                    self.pi_temp = self.pi_not
                else:
                    self.pi_not, self.alpha_not = fmm_gmm_init(
                        self.data, self.k, self.dist_combs[l])
                    self.alpha_temp = self.alpha_not
                    self.pi_temp = self.pi_not
                pi_new,alpha_new, log_likelihood_new,log_gamma_new=self._fit(self.data,self.pi_temp,self.alpha_temp,fmm_estimate_alphas,dist_comb=self.dist_combs[l])
                self.list_aic.append(
                    fmm_aic(alpha_new, log_likelihood_new, self.dist_combs[l] ))
                self.list_bic.append(
                    fmm_bic(alpha_new, log_likelihood_new,self.dist_combs[l], self.n))
                self.list_icl.append(
                    fmm_icl(alpha_new, log_likelihood_new,self.dist_combs[l], self.n, np.exp(log_gamma_new)))
                self.list_pi.append(pi_new)
                # self.list_alpha.append(convert_to_numpy(alpha_new))
                self.list_alpha.append(alpha_new)
                self.list_cluster.append(log_gamma_new.argmax(axis=1))
                self.list_log_likelihood.append(log_likelihood_new)
                self.list_gamma_matrix.append(np.exp(log_gamma_new))
                self.worked_dist.append(self.dist_combs[l])
                self.list_all_log_likelihood.append(self.log_likelihoods)
                if self.verbose:
                    print("distribution fitted", get_dist_names(self.dist_combs[l]))
            except:
                if self.debug:
                    traceback.print_exc()
                    print("Error received while running,",self.dist_combs[l])
                self.not_worked_dist.append(self.dist_combs[l])
                pass    
        self.list_pi,self.list_alpha,self.nan_ind=remove_nan_tuples(self.list_pi,self.list_alpha)
        self.list_aic=remove_elements_by_index(self.list_aic,self.nan_ind)
        self.list_bic=remove_elements_by_index(self.list_bic, self.nan_ind)
        self.list_icl=remove_elements_by_index(self.list_icl, self.nan_ind)
        self.list_cluster=remove_elements_by_index(self.list_cluster, self.nan_ind)
        self.list_log_likelihood=remove_elements_by_index(self.list_log_likelihood, self.nan_ind)
        self.list_gamma_matrix=remove_elements_by_index(self.list_gamma_matrix, self.nan_ind)
        self.not_worked_dist.extend(w for w in [self.worked_dist[h] for h in self.nan_ind])
        self.worked_dist=remove_elements_by_index(self.worked_dist, self.nan_ind)
        self.list_all_log_likelihood=remove_elements_by_index(self.list_all_log_likelihood, self.nan_ind)
        self.fitted=True
        if self.verbose:
            print("Model fitted successfully")
    
    def get_params(self):
        return self.list_pi, self.list_alpha

    def predict(self):
        return self.list_cluster

    def predict_new(self, X):
        # data_lol = x.values.tolist()
        cluster_all = []
        for l in range(len(self.dist_combs)):
            # cluster, _ = mixture_clusters(self.gamma_matrix[l], data_lol)
            cluster_all.append(self._predict(X))

        return cluster_all


    def best_mixture(self):
        best_mix = self.worked_dist[np.argmin(self.list_bic)]
        return [str(best_mix[i].__name__) for i in range(len(best_mix))]

    def best_params(self):
        return self.list_pi[np.argmin(self.list_aic)], self.list_alpha[np.argmin(self.list_aic)]

    def best_predict(self):
        return self.list_cluster[np.argmin(self.list_aic)]

    def best_predict_new(self, x):
        data_lol = x.values.tolist()
        cluster_all = []
        for l in range(len(self.dist_combs)):
            cluster, _ = mixture_clusters(self.gamma_matrix[l], data_lol)
            cluster_all.append(cluster)
        return cluster_all[np.argmin(self.list_aic)]

    def best_aic(self):
        return np.min(self.list_aic)

    def best_bic(self):
        return np.min(self.list_bic)
    def not_worked(self):
        print("Distribution Combinations That Could Not Be Fitted:")
        for i in range(len(self.not_worked_dist)):
            print(i, [str(self.not_worked_dist[i][j].__name__) for j in range(len(self.not_worked_dist[i]))])
    
    def worked(self):
        print("Distribution Combinations That Could Be Fitted:")
        for i in range(len(self.worked_dist)):
            print(i, [str(self.worked_dist[i][j].__name__) for j in range(len(self.worked_dist[i]))])
    
    def get_maximum_likelihood(self):
        return self.list_log_likelihood
    def get_all_likelihood(self):
        return self.list_all_log_likelihood
    
    
    def get_info_mat(self):
        """
        Assemble the total Fisher information matrix for each 'worked_dist' mixture
        distribution, combining the mixture-proportions info block and the
        parameter info blocks for each component.
    
        Returns:
          mixture_info_mats: list of 2D arrays, one per mixture distribution
          mixture_ses:       1D array of standard errors (concatenated across mixtures)
        """
        mixture_ses = []
        mixture_info_mats = []
    
        # Loop over each mixture distribution
        for a, mix in enumerate(self.worked_dist):
            # mix is the list of components for this mixture
            # e.g. mix = [comp_1, comp_2, ..., comp_k]
            k = len(mix)  # number of components
    
            # Extract data assigned to each component, etc.
            cwise_data = organize_data_by_clusters(np.array(self.data), self.list_cluster[a])
            mle_params = self.list_alpha[a]  # parameters for each component
            N_j, N = mixture_counts(self.list_gamma_matrix[a], mode="hard")
    
            # 1) Mixture proportion info
            I_pi, I_pi_inv = mixture_proportions_info(self.list_pi[a], N_j)
            # I_pi, I_pi_inv are k×k
    
            # 2) Component-level Fisher infos
            #    We'll collect the Fisher info for each component and store
            #    their inverses in I_inv_blocks for the final block assembly.
            mix_info = []
            param_dims = []    # dimension of each component's parameter vector
            I_inv_blocks = []
    
            for b, mix_dist in enumerate(mix):
                # Compute Fisher info for the b-th component
                fisher_info, param_len = compute_info_individual_fmvmm(
                    mix_dist,   # log PDF function
                    mle_params[b],     # parameters for that component
                    cwise_data[b],     # data assigned to that component
                    epsilon=0.06
                )
                mix_info.append(fisher_info)
                param_dims.append(param_len)
    
                # Invert it (or pseudo-invert if singular)
                try:
                    inv_block = np.linalg.inv(fisher_info)
                except np.linalg.LinAlgError:
                    inv_block = np.linalg.pinv(fisher_info)
    
                I_inv_blocks.append(inv_block)
    
            # 3) Now compute the total dimension:
            #    - k for the mixture proportions
            #    - plus the sum of parameter dims across all components
            total_param_dim = sum(param_dims)
            big_dim = k + total_param_dim
    
            # Allocate the overall info matrix and its inverse
            I_total = np.zeros((big_dim, big_dim))
            I_total_inv = np.zeros((big_dim, big_dim))
    
            # 4) Insert mixture proportion info in the top-left k×k block
            I_total[:k, :k] = I_pi
            I_total_inv[:k, :k] = I_pi_inv
    
            # 5) Insert each component's block
            row_start = k  # begin filling right below/after the k×k block
            for j in range(k):
                param_dim_j = param_dims[j]
                row_end = row_start + param_dim_j
    
                # Place the j-th component's fisher info block
                I_total[row_start:row_end, row_start:row_end] = mix_info[j]
                I_total_inv[row_start:row_end, row_start:row_end] = I_inv_blocks[j]
    
                row_start = row_end  # advance to the next block
    
            # 6) Compute standard errors from the diagonal of the total inverse
            #    (Ensure positive diagonals so we don't end up with sqrt of negative)
            var_diag = np.diag(ensure_positive_diagonal(I_total_inv))
            se_total = np.sqrt(var_diag)
    
            # 7) Store the results
            mixture_ses.append(se_total)      # collect standard errors
            mixture_info_mats.append(I_total) # collect the big Fisher matrix
    
        return mixture_info_mats, mixture_ses
    
    def get_top_mixtures(self,n_top=10):
        wrk_lst=[]
        for i in range(len(self.worked_dist)):
            wrk_lst.append([str(self.worked_dist[i][j].__name__) for j in range(len(self.worked_dist[i]))])

        # Combine the two lists into a list of tuples
        combined = list(zip(self.list_bic, wrk_lst))

        # Sort the list of tuples based on bic values
        sorted_combined = sorted(combined)

        return sorted_combined[:n_top]
                
                        


"""
Most of the codes are used from git repository mvem. Huge thanks to the
devoloper. 

"""

from scipy.stats import multivariate_normal
import numpy as np

def fit(X, return_loglike=False):
    """
    Estimate the parameters of the multivariate normal distribution using
    maximum likelihood estimates.

    :param x: An array of shape (n, p) containing n observations of some
        p-variate data with n > p.
    :type x: np.ndarray
    :param return_loglike: Return a list of log-likelihood values at each iteration. 
        Defaults to False.
    :type return_loglike: np.ndarray, optional
    :return: The fitted parameters (<array> mu, <array> sigma). Also returns a list of 
        log-likelihood values at each iteration of the EM algorithm if ``return_loglike=True``.
    :rtype: tuple
    """
    mean = np.mean(X, axis = 0)
    cov = np.cov(X.T)
    if return_loglike:
        return mean, cov, loglike(X, mean, cov)
    return mean, cov

def pdf(x, mean, cov, allow_singular=True):
    """
    Probability density function of the multivariate normal distribution.

    :param x: An array of shape (n, p) containing n observations of some
        p-variate data with n > p.
    :type x: np.ndarray
    :param mean: The mean of the normal distribution. A parameter with
        shape (p,).
    :type mean: np.ndarray
    :param cov: The covariance of the normal distribution. A positive
        semi-definite array with shape (p, p).
    :type cov: np.ndarray
    :param allow_singular: Whether to allow a singular matrix. Defaults to
        False.
    :type allow_singular: bool, optional
    :return: The density at each observation.
    :rtype: np.ndarray with shape (n,).        
    """
    return multivariate_normal.pdf(x, mean, cov, allow_singular)

def logpdf(x, mean, cov, allow_singular=True):
    """
    Log-probability density function of the multivariate normal distribution.

    :param x: An array of shape (n, p) containing n observations of some
        p-variate data with n > p.
    :type x: np.ndarray
    :param mean: The mean of the normal distribution. A parameter with
        shape (p,).
    :type mean: np.ndarray
    :param cov: The covariance of the normal distribution. A positive
        semi-definite array with shape (p, p).
    :type cov: np.ndarray
    :param allow_singular: Whether to allow a singular matrix. Defaults to
        False.
    :type allow_singular: bool, optional
    :return: The log-density at each observation.
    :rtype: np.ndarray with shape (n,).        
    """
    return multivariate_normal.logpdf(x, mean, cov, allow_singular)

def loglike(x, mean, cov, allow_singular=True):
    """
    Log-likelihood function of the multivariate normal distribution.

    :param x: An array of shape (n, p) containing n observations of some
        p-variate data with n > p.
    :type x: np.ndarray
    :param mean: The mean of the normal distribution. A parameter with
        shape (p,).
    :type mean: np.ndarray
    :param cov: The covariance of the normal distribution. A positive
        semi-definite array with shape (p, p).
    :type cov: np.ndarray
    :param allow_singular: Whether to allow a singular matrix. Defaults to
        False.
    :type allow_singular: bool, optional
    :return: The log-likelihood for given all observations and parameters.
    :rtype: float
    """
    return np.sum(logpdf(x, mean, cov, allow_singular))

def total_params(mean, cov):
    p = len(mean)
    
    return p + (p*(p+1)/2)

def cdf(x, mean, cov, allow_singular=True, maxpts=1000000, abseps=1e-5, releps=1e-5):
    """
    Cumulative density function of the multivariate normal distribution.

    :param x: An array of shape (n, p) containing n observations of some
        p-variate data with n > p.
    :type x: np.ndarray
    :param mean: The mean of the normal distribution. A parameter with
        shape (p,).
    :type mean: np.ndarray
    :param cov: The covariance of the normal distribution. A positive
        semi-definite array with shape (p, p).
    :type cov: np.ndarray
    :param allow_singular: Whether to allow a singular matrix. Defaults to
        False.
    :type allow_singular: bool, optional
    :param maxpts: The maximum number of points to use for integration.
    :type maxpts: int, optional
    :param abseps: The absolute error tolerance. Defaults to 1e-5.
    :type abseps: float, optional
    :param releps: The relative error tolerance. Defaults 1e-5.
    :type releps: float, optional
    :return: The cumulative density given all observations and parameters.
    :rtype: float
    """
    maxpts = maxpts * x.shape[1]
    return multivariate_normal.cdf(x, mean, cov, allow_singular, maxpts, abseps, releps)

def logcdf(x, mean, cov, allow_singular=True, maxpts=1000000, abseps=1e-5, releps=1e-5):
    """
    Log-cumulative density function of the multivariate normal distribution.

    :param x: An array of shape (n, p) containing n observations of some
        p-variate data with n > p.
    :type x: np.ndarray
    :param mean: The mean of the normal distribution. A parameter with
        shape (p,).
    :type mean: np.ndarray
    :param cov: The covariance of the normal distribution. A positive
        semi-definite array with shape (p, p).
    :type cov: np.ndarray
    :param allow_singular: Whether to allow a singular matrix. Defaults to
        False.
    :type allow_singular: bool, optional
    :param maxpts: The maximum number of points to use for integration.
    :type maxpts: int, optional
    :param abseps: The absolute error tolerance. Defaults to 1e-5.
    :type abseps: float, optional
    :param releps: The relative error tolerance. Defaults 1e-5.
    :type releps: float, optional
    :return: The log-cumulative density given all observations and parameters.
    :rtype: float
    """
    maxpts = maxpts * x.shape[1]
    return multivariate_normal.logcdf(x, mean, cov, allow_singular, maxpts, abseps, releps)

def rvs(mean, cov, size=1, random_state=None):
    """
    Random number generator of the multivariate normal distribution.

    :param mean: The mean of the normal distribution. A parameter with
        shape (p,).
    :type mean: np.ndarray
    :param cov: The covariance of the normal distribution. A positive
        semi-definite array with shape (p, p).
    :type cov: np.ndarray
    :param size: The number of samples to draw. Defaults to 1.
    :type size: int, optional
    :param random_state: Used for drawing random variates. Defaults to None.
    :type random_state: None, int, np.random.RandomState, np.random.Generator, optional
    :return: The random p-variate numbers generated.
    :rtype: np.ndarray with shape (n, p).       
    """
    return multivariate_normal.rvs(mean, cov, size, random_state)

def mean(mean, cov):
    """
    Mean function of the multivariate normal distribution.

    :param mean: The mean of the normal distribution. A parameter with
        shape (p,).
    :type mean: np.ndarray
    :param cov: The covariance of the normal distribution. A positive
        semi-definite array with shape (p, p).
    :type cov: np.ndarray
    :return: The mean of the specified distribution.
    :rtype: np.ndarray with shape (p,).
    """
    return mean

def var(mean, cov):
    """
    Variance function of the multivariate normal distribution.

    :param mean: The mean of the normal distribution. A parameter with
        shape (p,).
    :type mean: np.ndarray
    :param cov: The covariance of the normal distribution. A positive
        semi-definite array with shape (p, p).
    :type cov: np.ndarray
    :return: The variance of the specified distribution.
    :rtype: np.ndarray with shape (p,).
    """
    return cov



def info_mat(X, mu, Sigma):
    """
    Construct the Fisher information matrix for a multivariate normal's
    mean (p-dim) and full covariance (p^2-dim), ignoring the usual symmetry
    constraints on Sigma.

    Parameters
    ----------
    X : ndarray of shape (n, p)
        Observed data (not strictly needed for the 'expected' Fisher info,
        but we use n = X.shape[0]).
    mu : ndarray of shape (p,)
        The MLE of the mean (or any estimate at which you want the info).
    Sigma : ndarray of shape (p, p)
        The MLE of the covariance. Must be invertible.

    Returns
    -------
    I : ndarray of shape (p + p^2, p + p^2)
        The Fisher information block matrix with:
          - top-left : n * inv(Sigma), size p x p
          - bottom-right : (n/2) * (inv(Sigma) kronecker inv(Sigma)), size p^2 x p^2
          - cross-blocks = 0
        So total dimension is (p + p^2) x (p + p^2).

    Notes
    -----
    1) Ordinarily, Sigma is symmetric, so only p(p+1)/2 distinct parameters.
       Here we treat Sigma as if it has p^2 free entries. That will lead to
       singularities if you truly consider small variations that break symmetry.
    2) The 'observed' and 'expected' Fisher infos coincide for the normal at the MLE.
    """
    n, p = X.shape

    # Invert the MLE covariance
    inv_Sigma = np.linalg.inv(Sigma)

    # Block for mu: p x p
    I_mu = n * inv_Sigma

    # Block for Sigma: p^2 x p^2
    #   (n/2)*(inv(Sigma) kronecker inv(Sigma))
    kron_inv = np.kron(inv_Sigma, inv_Sigma)
    I_sigma = 0.5 * n * kron_inv

    # Form the full (p + p^2)x(p + p^2) block matrix
    I = np.zeros((p + p*p, p + p*p), dtype=float)

    # top-left: I_mu
    I[:p, :p] = I_mu
    # bottom-right: I_sigma
    I[p:, p:] = I_sigma

    return I



import numpy as np
# import numpy
# import cupy as np
import pandas as pd
from scipy.stats import dirichlet
import fmvmm.utils.dirichlet as drm
from fmvmm.utils.utils_dmm import (kmeans_init, gmm_init, kmeans_init_adv, gmm_init_adv, random_init,
                             dmm_loglikelihood, dmm_responsibilities, dmm_pi_estimate)


from fmvmm.utils.utils_mixture import (mixture_clusters)
import time
from scipy.special import gammaln,logsumexp
from scipy.special import polygamma, psi, digamma
import sys 
from fmvmm.mixtures._base import BaseMixture
from fmvmm.utils.utils_dmm import combined_info_and_se


MAXINT = sys.maxsize
euler = -1 * psi(1)  # Euler-Mascheroni constant



def estimate_alphas(data_cwise, alpha_not,dist_comb, method):
    alpha_new = []
    for t in range(len(data_cwise)):
        if np.array(data_cwise[t]).ndim == 1:
            alpha_new_temp = alpha_not[t]
        elif method == 'fixediteration':
            alpha_new_temp = drm.mle(
                np.array(data_cwise[t]), method="fixedpoint", maxiter=9223372036854775807)
        else:
            alpha_new_temp = drm.mle(np.array(data_cwise[t]))

        alpha_new.append(alpha_new_temp)
    return alpha_new





class DMM_Hard(BaseMixture):
    def __init__(self,n_clusters,tol=0.0001,initialization="kmeans",method="meanprecision",print_log_likelihood=False,max_iter=25, verbose=True):
        super().__init__(n_clusters=n_clusters, EM_type="Hard", mixture_type="identical", tol=tol, print_log_likelihood=print_log_likelihood, max_iter=max_iter, verbose=verbose)
        self.k=n_clusters
        self.initialization=initialization
        self.method = method
        self.family = "dirichlet"
        
    def _log_pdf_dirichlet(self,X,alphas):
        threshold = 1e-10
        N,p=X.shape
        k=len(alphas)
        probs=np.empty((N, k))
        for j in range(k):
            alpha=alphas[j]
            for i in range(N):
                with np.errstate(under="ignore"):
                    x=X[i, :]
                    t1=[(alphm-1)*np.log(xm) for alphm,xm in zip(alpha,x)]
                    # t1 = [(alphm - 1) * (np.log1p(xm - 1) if xm < threshold else np.log(xm)) for alphm, xm in zip(alpha, x)]
                    t2=np.sum(gammaln(alpha))
                    t3=gammaln(np.sum(alpha))
                    probs[i,j]=np.sum(t1)-t2+t3
        return probs    
    
    def _estimate_weighted_log_prob_identical(self, X, alpha, pi):
        return self._log_pdf_dirichlet(X,alpha) + np.log(pi)
    
    def fit(self,sample):
        start_time = time.time()
        self.data = self._process_data(sample)
        self.n, self.p = self.data.shape
        self.total_parameters = (self.k -1) + self.k* (self.p)
        np.random.seed(0)
        # if self.initialization == "kmeans":
        #     self.pi_not, self.alpha_not = kmeans_init(self.data, self.k)
        # elif self.initialization == "gmm":
        #     self.pi_not, self.alpha_not = gmm_init(self.data, self.k)
        if self.initialization == "kmeans":
            self.pi_not, self.alpha_not = kmeans_init_adv(self.data, self.k)
        elif self.initialization == "gmm":
            self.pi_not, self.alpha_not = gmm_init_adv(self.data, self.k)
        elif self.initialization == "random":
            self.pi_not, self.alpha_not = random_init(self.data, self.k)

        self.alpha_temp = self.alpha_not
        self.pi_temp = self.pi_not
        pi_new,alpha_new, log_likelihood_new,log_gamma_new=self._fit(self.data,self.pi_temp,self.alpha_temp,estimate_alphas,method=self.method)
        estimated_mean = []
        for a in alpha_new:
            mean_temp = [b/np.sum(a) for b in a]
            estimated_mean.append(mean_temp)
        self.pi_new = pi_new
        self.alpha_new = alpha_new
        self.estimated_mean = estimated_mean
        self.cluster = log_gamma_new.argmax(axis=1)
        # self.data_cwise = data_cwise
        self.gamma_temp_ar = np.exp(log_gamma_new)
        # self.gamma_matrix = gamma_matrix
        self.log_likelihood_new = log_likelihood_new
        if self.verbose:
            print("Hard DMM Fitting Done Successfully")
        end_time = time.time()
        self.execution_time=end_time-start_time
    
    
    def get_params(self):
        #print("The estimated pi values are ", self.pi_new)
        #print("The estimated alpha values are ", self.alpha_new)

        return self.pi_new, self.alpha_new

    def predict(self):
        return self.cluster

    def predict_new(self, x):
        x=self._process_data(x)
        data_lol = x.tolist()
        cluster, _ = mixture_clusters(self.gamma_temp_ar, data_lol)

        return cluster

    def get_mean(self):

        return self.estimated_mean

    def get_precision(self):
        preci=[np.sum(al) for al in self.alpha_new]
        
        return preci

    def responsibilities(self):

        return self.gamma_temp_ar

    def n_iter(self):
        return len(self.log_likelihoods)
    
    def get_info_mat(self):
        IM, SE = combined_info_and_se(self.pi_new, np.array(self.alpha_new), self.gamma_temp_ar, mode = "hard")
        
        return IM, SE

    # def clustered_data(self):
    #     return pd.DataFrame(self.data_cwise).transpose()

    # def bic(self):

    #     return ((self.k-1)+(self.k*self.p))*np.log(self.n) - 2*(self.log_likelihood_new)

    # def aic(self):

    #     return 2*((self.k-1)+(self.k*self.p)) - 2*(self.log_likelihood_new)

    # def icl(self):
    #     entropy_s=[]
    #     for i in self.gamma_matrix:
    #         for j in i:
    #                 ent_temp=j*np.log1p(j)
    #                 entropy_s.append(ent_temp)



    #     entropy=np.sum(entropy_s)

    #     return ((self.k-1)+(self.k*self.p))*np.log(self.n) - 2*(self.log_likelihood_new) - entropy

    


import numpy as np
from scipy.stats import gamma, norm, invgamma, invgauss
from scipy.optimize import root_scalar

def gig_y_gfn(y, m, beta, lambda_):
    y2 = y * y
    g = 0.5 * beta * y2 * y
    g -= y2 * (0.5 * beta * m + lambda_ + 1.0)
    g += y * ((lambda_ - 1.0) * m - 0.5 * beta) + 0.5 * beta * m
    return g

def rinvgauss(mu, lambda_):
    y = norm.rvs()
    y *= y
    mu2 = mu * mu
    l2 = 2.0 * lambda_
    x1 = mu + mu2 * y / l2 - (mu / l2) * np.sqrt(4.0 * mu * lambda_ * y + mu2 * y * y)

    u = np.random.uniform()
    if u <= mu / (mu + x1):
        return x1
    else:
        return mu2 / x1

def zeroin_gig(ax, bx, f, tol, m, beta, lambda_):
    a, b, c = ax, bx, ax
    fa, fb, fc = f(a, m, beta, lambda_), f(b, m, beta, lambda_), f(c, m, beta, lambda_)

    for _ in range(1000):  # Maximum iterations to avoid infinite loops
        prev_step = b - a
        tol_act = 2.0 * np.sqrt(np.finfo(float).eps) * np.abs(b) + tol / 2.0
        new_step = (c - b) / 2.0

        if np.abs(new_step) <= tol_act or fb == 0:
            return b

        if np.abs(prev_step) >= tol_act and np.abs(fa) > np.abs(fb):
            cb = c - b

            if a == c:
                t1 = fb / fa
                p = cb * t1
                q = 1.0 - t1
            else:
                q = fa / fc
                t1 = fb / fc
                t2 = fb / fa
                p = t2 * (cb * q * (q - t1) - (b - a) * (t1 - 1.0))
                q = (q - 1.0) * (t1 - 1.0) * (t2 - 1.0)

            if p > 0:
                q = -q
            else:
                p = -p

            if p < (0.75 * cb * q - np.abs(tol_act * q) / 2.0) and p < np.abs(prev_step * q / 2.0):
                new_step = p / q

        if np.abs(new_step) < tol_act:
            if new_step > 0:
                new_step = tol_act
            else:
                new_step = -tol_act

        a, fa = b, fb
        b += new_step
        fb = f(b, m, beta, lambda_)

        if (fb > 0 and fc > 0) or (fb < 0 and fc < 0):
            c, fc = a, fa

def rgig(n, lambda_, chi, psi):
    ZTOL = np.sqrt(np.finfo(float).eps)
    samps = np.empty(n)

    if chi < ZTOL and lambda_ > 0.0:
        samps = gamma.rvs(lambda_, scale=2.0/psi, size=n)
        return samps

    if psi < ZTOL and lambda_ < 0.0:
        samps = 1.0 / gamma.rvs(0.0 - lambda_, scale=2.0/chi, size=n)
        return samps

    if lambda_ == -0.5:
        alpha = np.sqrt(chi / psi)
        samps = invgauss.rvs(alpha, scale=chi, size=n)
        return samps

    alpha = np.sqrt(chi / psi)
    beta2 = psi * chi
    beta = np.sqrt(psi * chi)
    lm1 = lambda_ - 1.0
    lm12 = lm1 * lm1
    m = (lm1 + np.sqrt(lm12 + beta2)) / beta
    m1 = m + 1.0/m

    upper = m
    while gig_y_gfn(upper, m, beta, lambda_) <= 0:
        upper *= 2.0

    yM = zeroin_gig(0.0, m, gig_y_gfn, ZTOL, m, beta, lambda_)
    yP = zeroin_gig(m, upper, gig_y_gfn, ZTOL, m, beta, lambda_)

    a = (yP - m) * (yP/m) ** (0.5 * lm1)
    a *= np.exp(-0.25 * beta * (yP + 1.0/yP - m1))
    b = (yM - m) * (yM/m) ** (0.5 * lm1)
    b *= np.exp(-0.25 * beta * (yM + 1.0/yM - m1))
    c = -0.25 * beta * m1 + 0.5 * lm1 * np.log(m)

    need = 1
    for i in range(n):
        while need:
            R1, R2 = np.random.uniform(size=2)
            Y = m + a * R2 / R1 + b * (1.0 - R2) / R1
            if Y > 0.0:
                if -np.log(R1) >= -0.5 * lm1 * np.log(Y) + 0.25 * beta * (Y + 1.0/Y) + c:
                    need = 0
        samps[i] = Y * alpha
        need = 1

    return samps
